@@ -41,7 +41,12 @@ NASC_ARGUMENTS = {
     '--scoreInsOpen': -10000,
 }
 
-BAM_PEEK_READS = 200000
+BAM_PEEK_READS = 500000
 BAM_REQUIRED_TAGS = ['MD']
-COUNTS_SPLIT_THRESHOLD = 20000
+BAM_READGROUP_TAG = 'RG'
+BAM_BARCODE_TAG = 'CB'
+BAM_UMI_TAG = 'UB'
+BAM_GENE_TAG = 'GX'
+BAM_CONSENSUS_READ_COUNT_TAG = 'RN'
+COUNTS_SPLIT_THRESHOLD = 50000
 VELOCITY_BLACKLIST = ['unassigned', 'ambiguous']
