@@ -11,6 +11,8 @@ STAR_BARCODES_FILENAME = 'barcodes.tsv'
 STAR_FEATURES_FILENAME = 'features.tsv'
 STAR_MATRIX_FILENAME = 'matrix.mtx'
 
+CONSENSUS_BAM_FILENAME = 'consensus.bam'
+
 COUNT_DIR = 'count'
 PARSE_DIR = '0_parse'
 CONVS_FILENAME = 'convs.pkl.gz'
@@ -24,9 +26,8 @@ COVERAGE_INDEX_FILENAME = 'coverage.idx'
 SNPS_FILENAME = 'snps.csv'
 
 COUNTS_PREFIX = 'counts'
-
 ESTIMATE_DIR = 'estimate'
-RATES_FILENAME = 'rates.csv'
+RATES_PREFIX = 'rates'
 P_E_FILENAME = 'p_e.csv'
 P_C_PREFIX = 'p_c'
 AGGREGATE_FILENAME = 'aggregate.csv'
