@@ -4,19 +4,21 @@ import os
 import shutil
 import sys
 import warnings
+from typing import Optional
 
 from . import __version__
-# from .config import RE_CHOICES
+from .config import BAM_GENE_TAG
 from .logging import logger
-from .technology import TECHNOLOGIES_MAP
-from .utils import flatten_list, patch_mp_connection_bpo_17560
+from .preprocessing import CONVERSION_COMPLEMENT
+from .technology import STRAND_MAP, TECHNOLOGIES_MAP, detect_strand
+from .utils import flatten_iter, patch_mp_connection_bpo_17560
 
 
 def print_technologies():
     """Displays a list of supported technologies along with whether a whitelist
     is provided for that technology.
     """
-    headers = ['name', 'whitelist', 'barcode', 'umi', 'cDNA']
+    headers = ['name', 'whitelist', 'barcode', 'umi', 'cDNA', 'strand']
     rows = [headers]
 
     print('List of supported single-cell technologies\n')
@@ -26,11 +28,10 @@ def print_technologies():
         t = TECHNOLOGIES_MAP[key]
         chem = t.chemistry
         row = [
-            t.name,
-            'yes' if chem.has_whitelist else '',
+            t.name, 'yes' if chem.has_whitelist else '',
             ' '.join(str(_def) for _def in chem.cell_barcode_parser) if chem.has_cell_barcode else '',
             ' '.join(str(_def) for _def in chem.umi_parser) if chem.has_umi else '',
-            ' '.join(str(_def) for _def in chem.cdna_parser),
+            ' '.join(str(_def) for _def in chem.cdna_parser), chem.strand.name
         ]
         rows.append(row)
 
@@ -48,17 +49,16 @@ def print_technologies():
     sys.exit(1)
 
 
-def setup_ref_args(parser, parent):
+def setup_ref_args(parser: argparse.ArgumentParser, parent: argparse.ArgumentParser) -> argparse.ArgumentParser:
     """Helper function to set up a subparser for the `ref` command.
 
-    :param parser: argparse parser to add the `ref` command to
-    :type args: argparse.ArgumentParser
-    :param parent: argparse parser parent of the newly added subcommand.
-                   used to inherit shared commands/flags
-    :type args: argparse.ArgumentParser
+    Args:
+        parser: Argparse parser to add the `ref` command to
+        parent: Argparse parser parent of the newly added subcommand.
+            Used to inherit shared commands/flags
 
-    :return: the newly added parser
-    :rtype: argparse.ArgumentParser
+    Returns:
+        The newly added parser
     """
     parser_ref = parser.add_parser(
         'ref',
@@ -93,7 +93,17 @@ def setup_ref_args(parser, parent):
     return parser_ref
 
 
-def setup_align_args(parser, parent):
+def setup_align_args(parser: argparse.ArgumentParser, parent: argparse.ArgumentParser) -> argparse.ArgumentParser:
+    """Helper function to set up a subparser for the `align` command.
+
+    Args:
+        parser: Argparse parser to add the `align` command to
+        parent: Argparse parser parent of the newly added subcommand.
+            Used to inherit shared commands/flags
+
+    Returns:
+        The newly added parser
+    """
     parser_align = parser.add_parser(
         'align',
         description='Align FASTQs',
@@ -123,9 +133,9 @@ def setup_align_args(parser, parent):
     )
     parser_align.add_argument(
         '--strand',
-        help='Read strandedness. (default: `forward`)',
-        choices=['forward', 'reverse', 'unstranded'],
-        default='forward',
+        help='Read strandedness. By default, this is auto-detected depending on the technology.',
+        choices=['forward', 'reverse', 'unstranded', 'auto'],
+        default='auto',
     )
     parser_align.add_argument(
         '-w',
@@ -133,7 +143,7 @@ def setup_align_args(parser, parent):
         help=('Path to file of whitelisted barcodes to correct to. '
               'If not provided, all barcodes are used.'),
     )
-    parser_align.add_argument('--overwrite', help='Overwrite existing alignment files', action='store_true')
+    parser_align.add_argument('--overwrite', help='Overwrite existing files.', action='store_true')
     parser_align.add_argument(
         '--STAR-overrides', metavar='ARGUMENTS', help='Arguments to pass directly to STAR.', type=str, default=None
     )
@@ -156,17 +166,119 @@ def setup_align_args(parser, parent):
     return parser_align
 
 
-def setup_count_args(parser, parent):
+def setup_consensus_args(parser: argparse.ArgumentParser, parent: argparse.ArgumentParser) -> argparse.ArgumentParser:
+    """Helper function to set up a subparser for the `consensus` command.
+
+    Args:
+        parser: Argparse parser to add the `consensus` command to
+        parent: Argparse parser parent of the newly added subcommand.
+            Used to inherit shared commands/flags
+
+    Returns:
+        The newly added parser
+    """
+    parser_consensus = parser.add_parser(
+        'consensus',
+        description='Generate consensus sequences',
+        help='Generate consensus sequences',
+        parents=[parent],
+    )
+    parser_consensus._actions[0].help = parser_consensus._actions[0].help.capitalize()
+
+    required_consensus = parser_consensus.add_argument_group('required arguments')
+    required_consensus.add_argument(
+        '-g', metavar='GTF', help='Path to GTF file used to generate the STAR index', type=str, required=True
+    )
+    parser_consensus.add_argument(
+        '-o',
+        metavar='OUT',
+        help='Path to output directory (default: current directory)',
+        type=str,
+        default='.',
+    )
+    parser_consensus.add_argument(
+        '--umi-tag',
+        metavar='TAG',
+        help=(
+            'BAM tag to use as unique molecular identifiers (UMI). If not provided, '
+            'all reads are assumed to be unique. (default: None)'
+        ),
+        type=str,
+        default=None,
+    )
+    parser_consensus.add_argument(
+        '--barcode-tag',
+        metavar='TAG',
+        help=(
+            'BAM tag to use as cell barcodes. If not provided, all reads are '
+            'assumed to be from a single cell. (default: None)'
+        ),
+        type=str,
+        default=None,
+    )
+    parser_consensus.add_argument(
+        '--gene-tag',
+        metavar='TAG',
+        help=f'BAM tag to use as gene assignments (default: {BAM_GENE_TAG})',
+        type=str,
+        default=BAM_GENE_TAG,
+    )
+    parser_consensus.add_argument(
+        '--strand',
+        help='Read strandedness. By default, this is auto-detected from the BAM.',
+        choices=['forward', 'reverse', 'unstranded', 'auto'],
+        default='auto',
+    )
+    parser_consensus.add_argument(
+        '--quality',
+        metavar='QUALITY',
+        help=(
+            'Base quality threshold. When generating a consensus nucleotide at a '
+            'certain position, the base with smallest error probability below this '
+            'quality threshold is chosen. If no base meets this criteria, the '
+            'reference base is chosen. (default: 27)'
+        ),
+        type=int,
+        default=27
+    )
+    parser_consensus.add_argument(
+        '--barcodes',
+        metavar='TXT',
+        help=('Textfile containing filtered cell barcodes. Only these barcodes will '
+              'be processed.'),
+        type=str,
+    )
+    parser_consensus.add_argument(
+        '--add-RS-RI',
+        help=(
+            'Add custom RS and RI tags to the output BAM, each of which contain a '
+            'semi-colon delimited list of read names (RS) and alignment indices (RI) '
+            'of the reads and alignments from which the consensus is derived. This '
+            'option is useful for debugging.'
+        ),
+        action='store_true'
+    )
+    parser_consensus.add_argument(
+        'bam',
+        help=(
+            'Alignment BAM file that contains the appropriate UMI and barcode tags, '
+            'specifiable with `--umi-tag`, and `--barcode-tag`.'
+        ),
+        type=str,
+    )
+    return parser_consensus
+
+
+def setup_count_args(parser: argparse.ArgumentParser, parent: argparse.ArgumentParser) -> argparse.ArgumentParser:
     """Helper function to set up a subparser for the `count` command.
 
-    :param parser: argparse parser to add the `count` command to
-    :type args: argparse.ArgumentParser
-    :param parent: argparse parser parent of the newly added subcommand.
-                   used to inherit shared commands/flags
-    :type args: argparse.ArgumentParser
+    Args:
+        parser: Argparse parser to add the `count` command to
+        parent: Argparse parser parent of the newly added subcommand.
+            Used to inherit shared commands/flags
 
-    :return: the newly added parser
-    :rtype: argparse.ArgumentParser
+    Returns:
+        The newly added parser
     """
     parser_count = parser.add_parser(
         'count',
@@ -190,7 +302,8 @@ def setup_count_args(parser, parent):
             'timepoint.'
         ),
         action='append',
-        required=True
+        required=True,
+        choices=sorted(CONVERSION_COMPLEMENT.keys())
     )
     parser_count.add_argument(
         '-o',
@@ -222,15 +335,15 @@ def setup_count_args(parser, parent):
     parser_count.add_argument(
         '--gene-tag',
         metavar='TAG',
-        help='BAM tag to use as gene assignments (default: GX)',
+        help=f'BAM tag to use as gene assignments (default: {BAM_GENE_TAG})',
         type=str,
-        default='GX',
+        default=BAM_GENE_TAG,
     )
     parser_count.add_argument(
         '--strand',
-        help='Read strandedness. (default: `forward`)',
-        choices=['forward', 'reverse', 'unstranded'],
-        default='forward',
+        help='Read strandedness. By default, this is auto-detected from the BAM.',
+        choices=['forward', 'reverse', 'unstranded', 'auto'],
+        default='auto',
     )
     parser_count.add_argument(
         '--quality',
@@ -251,7 +364,17 @@ def setup_count_args(parser, parent):
             '(default: no SNP detection)'
         ),
         type=float,
-        default=False,
+        default=None,
+    )
+    parser_count.add_argument(
+        '--snp-min-coverage',
+        metavar='THRESHOLD',
+        help=(
+            'For a conversion to be considered as a SNP, there must be at least '
+            'this many reads mapping to that region. (default: 1)'
+        ),
+        type=int,
+        default=1,
     )
     parser_count.add_argument(
         '--snp-csv',
@@ -269,6 +392,13 @@ def setup_count_args(parser, parent):
         type=str,
     )
     parser_count.add_argument(
+        '--gene-names',
+        help=('Group counts by gene names instead of gene IDs when generating the h5ad file.'),
+        action='store_true'
+    )
+
+    splicing_group = parser_count.add_mutually_exclusive_group()
+    splicing_group.add_argument(
         '--no-splicing',
         '--transcriptome-only',
         help=(
@@ -277,6 +407,18 @@ def setup_count_args(parser, parent):
         ),
         action='store_true'
     )
+    splicing_group.add_argument(
+        '--exon-overlap',
+        help=(
+            'Algorithm to use to detect spliced reads (that overlap exons). '
+            'May be `strict`, which assigns reads as spliced if it only overlaps '
+            'exons, or `lenient`, which assigns reads as spliced if it does not '
+            'overlap with any introns of at least one transcript. (default: strict)'
+        ),
+        type=str,
+        choices=['lenient', 'strict'],
+        default='strict',
+    )
     parser_count.add_argument(
         '--nasc',
         help=argparse.SUPPRESS,
@@ -287,9 +429,21 @@ def setup_count_args(parser, parent):
         help='Indicate this is a control sample, which is used to detect SNPs.',
         action='store_true',
     )
+    parser_count.add_argument(
+        '--dedup-mode',
+        help=(
+            'Deduplication mode for UMI-based technologies (required `--umi-tag`). '
+            'Available choices are: `auto`, `conversion`, `exon`. When `conversion` is used, '
+            'reads that have at least one of the provided conversions is prioritized. '
+            'When `exon` is used, exonic reads are prioritized. By default (`auto`), '
+            'the BAM is inspected to select the appropriate mode.'
+        ),
+        type=str,
+        choices=['auto', 'conversion', 'exon'],
+    )
     parser_count.add_argument(
         '--overwrite',
-        help='',
+        help='Overwrite existing files.',
         action='store_true',
     )
     parser_count.add_argument(
@@ -304,17 +458,16 @@ def setup_count_args(parser, parent):
     return parser_count
 
 
-def setup_estimate_args(parser, parent):
+def setup_estimate_args(parser: argparse.ArgumentParser, parent: argparse.ArgumentParser) -> argparse.ArgumentParser:
     """Helper function to set up a subparser for the `estimate` command.
 
-    :param parser: argparse parser to add the `estimate` command to
-    :type args: argparse.ArgumentParser
-    :param parent: argparse parser parent of the newly added subcommand.
-                   used to inherit shared commands/flags
-    :type args: argparse.ArgumentParser
+    Args:
+        parser: Argparse parser to add the `estimate` command to
+        parent: Argparse parser parent of the newly added subcommand.
+            Used to inherit shared commands/flags
 
-    :return: the newly added parser
-    :rtype: argparse.ArgumentParser
+    Returns:
+        The newly added parser
     """
     parser_estimate = parser.add_parser(
         'estimate',
@@ -344,20 +497,44 @@ def setup_estimate_args(parser, parent):
         default='.',
     )
     parser_estimate.add_argument(
-        '--groups',
-        metavar='CSV',
+        '--barcodes',
+        metavar='TXT',
         help=(
-            'CSV containing cell (barcode) groups, where the first column is the barcode '
-            'and the second is the group name the cell belongs to. Cells will be combined per '
-            'group for estimation of parameters specified by `--groups-for`.'
+            'Textfile containing filtered cell barcodes. Only these barcodes will be processed. '
+            'This option may be used multiple times when multiple input directories are provided.'
         ),
         action='append',
         default=None,
     )
     parser_estimate.add_argument(
-        '--ignore-groups-for-pi',
+        '--groups',
+        metavar='CSV',
         help=(
-            'Ignore cell groupings when estimating the fraction of labeled RNA. '
+            'CSV containing cell (barcode) groups, where the first column is the barcode '
+            'and the second is the group name the cell belongs to. Estimation will be performed '
+            'by aggregating UMIs per group. This option may be used multiple times when multiple '
+            'input directories are provided.'
+        ),
+        action='append',
+        default=None,
+    )
+    parser_estimate.add_argument(
+        '--method',
+        help=(
+            'Correction method to use. '
+            'May be `pi_g` to estimate the fraction of labeled RNA for every cell-gene combination, '
+            'or `alpha` to use alpha correction as used in the scNT-seq paper. `alpha` is recommended for '
+            'UMI-based assays. This option has no effect when used with `--control`. (default: alpha)'
+        ),
+        choices=['pi_g', 'alpha'],
+        default='alpha'
+    )
+    parser_estimate.add_argument(
+        '--ignore-groups-for-est',
+        help=(
+            'Ignore cell groupings when calculating final estimations for the fraction of labeled RNA. '
+            'When `--method pi_g`, groups are ignored when estimating fraction of labeled RNA. '
+            'When `--method alpha`, groups are ignored when estimating detection rate. '
             'This option only has an effect when `--groups` is also specified.'
         ),
         action='store_true',
@@ -379,10 +556,18 @@ def setup_estimate_args(parser, parent):
     parser_estimate.add_argument(
         '--cell-gene-threshold',
         metavar='COUNT',
-        help='A cell-gene pair must have at least this many reads for correction. (default: 16)',
+        help=(
+            'A cell-gene pair must have at least this many reads for correction. '
+            'Only for `--method pi_g`. (default: 16)'
+        ),
         type=int,
         default=16
     )
+    parser_estimate.add_argument(
+        '--gene-names',
+        help=('Group counts by gene names instead of gene IDs when generating H5AD file'),
+        action='store_true'
+    )
     parser_estimate.add_argument(
         '--downsample',
         metavar='NUM',
@@ -439,10 +624,13 @@ def setup_estimate_args(parser, parent):
     return parser_estimate
 
 
-def parse_ref(parser, args, temp_dir=None):
+def parse_ref(parser: argparse.ArgumentParser, args: argparse.Namespace, temp_dir: Optional[str] = None):
     """Parser for the `ref` command.
-    :param args: Command-line arguments dictionary, as parsed by argparse
-    :type args: dict
+
+    Args:
+        parser: The parser
+        args: Command-line arguments dictionary, as parsed by argparse
+        temp_dir: Temporary directory
     """
     if os.path.exists(args.i):
         parser.error(
@@ -514,6 +702,15 @@ def parse_align(parser, args, temp_dir=None):
         # Clean
         overrides = {arg: parts[0] if len(parts) == 1 else parts for arg, parts in overrides.items()}
 
+    # Detect strand
+    strand = args.strand
+    if args.strand == 'auto':
+        strand = STRAND_MAP[technology.chemistry.strand]
+        logger.info(
+            f'Auto-detected strandedness `{strand}` for technology `{technology.name}`. '
+            'Use `--strand` to override.'
+        )
+
     from .align import align
     align(
         args.fastqs,
@@ -521,7 +718,7 @@ def parse_align(parser, args, temp_dir=None):
         args.o,
         technology,
         whitelist_path=args.w,
-        strand=args.strand,
+        strand=strand,
         n_threads=args.t,
         temp_dir=temp_dir,
         nasc=args.nasc,
@@ -529,10 +726,13 @@ def parse_align(parser, args, temp_dir=None):
     )
 
 
-def parse_count(parser, args, temp_dir=None):
-    """Parser for the `count` command.
-    :param args: Command-line arguments dictionary, as parsed by argparse
-    :type args: dict
+def parse_consensus(parser: argparse.ArgumentParser, args: argparse.Namespace, temp_dir: Optional[str] = None):
+    """Parser for the `consensus` command.
+
+    Args:
+        parser: The parser
+        args: Command-line arguments dictionary, as parsed by argparse
+        temp_dir: Temporary directory
     """
     # Check quality
     if args.quality < 0 or args.quality > 41:
@@ -541,8 +741,61 @@ def parse_count(parser, args, temp_dir=None):
     # Read barcodes
     barcodes = None
     if args.barcodes:
+        if not args.barcode_tag:
+            parser.error('`--barcodes` may only be provided with `--barcode-tag`.')
+
         with open(args.barcodes, 'r') as f:
-            barcodes = [line.strip() for line in f if not line.isspace()]
+            barcodes = set(line.strip() for line in f if not line.isspace())
+        logger.warning(f'Ignoring cell barcodes not in the {len(barcodes)} barcodes provided by `--barcodes`')
+    else:
+        logger.warning('`--barcodes` not provided. All cell barcodes will be processed.')
+
+    # Detect strand
+    strand = args.strand
+    if args.strand == 'auto':
+        strand = detect_strand(args.bam)
+        if strand is None:
+            parser.error('Failed to auto-detect strandedness from BAM. Use `--strand` to provide strandedness.')
+        else:
+            logger.info(f'Auto-detected strandedness: {strand}. Use `--strand` to override.')
+
+    from .consensus import consensus
+    consensus(
+        args.bam,
+        args.g,
+        args.o,
+        strand=strand,
+        umi_tag=args.umi_tag,
+        barcode_tag=args.barcode_tag,
+        gene_tag=args.gene_tag,
+        barcodes=barcodes,
+        quality=args.quality,
+        add_RS_RI=args.add_RS_RI,
+        n_threads=args.t,
+        temp_dir=temp_dir,
+    )
+
+
+def parse_count(parser: argparse.ArgumentParser, args: argparse.Namespace, temp_dir: Optional[str] = None):
+    """Parser for the `count` command.
+
+    Args:
+        parser: The parser
+        args: Command-line arguments dictionary, as parsed by argparse
+        temp_dir: Temporary directory
+    """
+    # Check quality
+    if args.quality < 0 or args.quality > 41:
+        parser.error('`--quality` must be in [0, 42)')
+
+    # Read barcodes
+    barcodes = None
+    if args.barcodes:
+        if not args.barcode_tag:
+            parser.error('`--barcodes` may only be provided with `--barcode-tag`.')
+
+        with open(args.barcodes, 'r') as f:
+            barcodes = set(line.strip() for line in f if not line.isspace())
         logger.warning(f'Ignoring cell barcodes not in the {len(barcodes)} barcodes provided by `--barcodes`')
     else:
         logger.warning('`--barcodes` not provided. All cell barcodes will be processed.')
@@ -556,16 +809,28 @@ def parse_count(parser, args, temp_dir=None):
     conversions = []
     for arg in args.conversion:
         conversions.append(arg.upper().split(','))
-    flattened = list(flatten_list(conversions))
+    flattened = list(flatten_iter(conversions))
     if len(set(flattened)) != len(flattened):
         parser.error('duplicate conversions are not allowed for `--conversion`')
 
+    # Convert conversions to frozenset of tuples.
+    conversions = frozenset(frozenset(conv) for conv in conversions)
+
+    # Detect strand
+    strand = args.strand
+    if args.strand == 'auto':
+        strand = detect_strand(args.bam)
+        if strand is None:
+            parser.error('Failed to auto-detect strandedness from BAM. Use `--strand` to provide strandedness.')
+        else:
+            logger.info(f'Auto-detected strandedness: {strand}. Use `--strand` to override.')
+
     from .count import count
     count(
         args.bam,
         args.g,
         args.o,
-        strand=args.strand,
+        strand=strand,
         umi_tag=args.umi_tag,
         barcode_tag=args.barcode_tag,
         gene_tag=args.gene_tag,
@@ -574,16 +839,27 @@ def parse_count(parser, args, temp_dir=None):
         quality=args.quality,
         conversions=conversions,
         snp_threshold=args.snp_threshold,
+        snp_min_coverage=args.snp_min_coverage,
         snp_csv=args.snp_csv,
         n_threads=args.t,
         temp_dir=temp_dir,
         nasc=args.nasc,
         overwrite=args.overwrite,
         velocity=not args.no_splicing,
+        strict_exon_overlap=args.exon_overlap == 'strict',
+        dedup_mode=args.dedup_mode,
+        by_name=args.gene_names
     )
 
 
-def parse_estimate(parser, args, temp_dir=None):
+def parse_estimate(parser: argparse.ArgumentParser, args: argparse.Namespace, temp_dir: Optional[str] = None):
+    """Parser for the `estimate` command.
+
+    Args:
+        parser: The parser
+        args: Command-line arguments dictionary, as parsed by argparse
+        temp_dir: Temporary directory
+    """
     # Check control constraints
     if args.control and args.p_e:
         parser.error('`--control` and `--p-e` can not be used together')
@@ -599,11 +875,16 @@ def parse_estimate(parser, args, temp_dir=None):
 
     # group CSV must be proivded when there are multiple input directories
     if len(args.count_dirs) > 1 and not args.groups:
-        parser.error('`--group` CSVs must be provided when using multiple input directories')
+        parser.error(
+            '`--group` CSVs must be provided when using multiple input directories. '
+            'Otherwise, simply run `dynast estimate` for each directory separately.'
+        )
 
     # If group CSV(s) are provided, the number must match input directories
     if args.groups and len(args.groups) != len(args.count_dirs):
         parser.error('Number of `--group` CSVs must match number of input directories')
+    if args.barcodes and len(args.barcodes) != len(args.count_dirs):
+        parser.error('Number of `--barcodes` TXTs must match number of input directories')
 
     # Multiple count dirs can't be used with nasc
     if len(args.count_dirs) > 1 and args.nasc:
@@ -612,13 +893,36 @@ def parse_estimate(parser, args, temp_dir=None):
     # Read genes
     genes = None
     if args.genes:
+        if args.by_name:
+            logger.warning(
+                '`--genes` were provided with `--gene-names`. '
+                'Make sure your gene list contains gene names instead of IDs. '
+                'IDs should be used for any genes that do not have a name.'
+            )
         with open(args.genes, 'r') as f:
             genes = [line.strip() for line in f if not line.isspace()]
         logger.warning(f'Ignoring genes not in the {len(genes)} genes provided by `--genes`')
 
+    barcodes = []
+    if args.barcodes:
+        for path in args.barcodes:
+            with open(path, 'r') as f:
+                bcs = set(line.strip() for line in f if not line.isspace())
+            barcodes.append(bcs)
+        logger.warning(
+            f'Ignoring cell barcodes not in the {sum(len(bcs) for bcs in barcodes)} barcodes provided by `--barcodes`'
+        )
+    else:
+        logger.warning('`--barcodes` not provided. All cell barcodes will be processed.')
+
     # Parse cell groups csv(s)
     groups = []
     if args.groups:
+        if args.barcodes:
+            logger.warning(
+                '`--groups` was provided with `--barcodes`. Only barcodes present in both inputs will be considered.'
+            )
+
         for path in args.groups:
             groups_part = {}
             with open(path, 'r') as f:
@@ -633,7 +937,7 @@ def parse_estimate(parser, args, temp_dir=None):
                     groups_part[barcode] = group
             groups.append(groups_part)
 
-    if args.ignore_groups_for_pi and not (args.groups):
+    if args.ignore_groups_for_est and not (args.groups):
         parser.error('`--ignore-groups-for-pi` can not be used without `--groups`')
 
     if not args.reads:
@@ -651,8 +955,9 @@ def parse_estimate(parser, args, temp_dir=None):
         args.count_dirs,
         args.o,
         args.reads,
+        barcodes=args.barcodes,
         groups=groups,
-        ignore_groups_for_pi=args.ignore_groups_for_pi,
+        ignore_groups_for_est=args.ignore_groups_for_est,
         genes=genes,
         downsample=args.downsample,
         downsample_mode=args.downsample_mode,
@@ -660,9 +965,11 @@ def parse_estimate(parser, args, temp_dir=None):
         cell_gene_threshold=args.cell_gene_threshold,
         control_p_e=control_p_e,
         control=args.control,
+        method=args.method,
         n_threads=args.t,
         temp_dir=temp_dir,
         nasc=args.nasc,
+        by_name=args.gene_names,
         seed=args.seed
     )
 
@@ -670,6 +977,7 @@ def parse_estimate(parser, args, temp_dir=None):
 COMMAND_TO_FUNCTION = {
     'ref': parse_ref,
     'align': parse_align,
+    'consensus': parse_consensus,
     'count': parse_count,
     'estimate': parse_estimate,
 }
@@ -697,11 +1005,13 @@ def main():
     # Command parsers
     parser_ref = setup_ref_args(subparsers, parent)
     parser_align = setup_align_args(subparsers, parent)
+    parser_consensus = setup_consensus_args(subparsers, parent)
     parser_count = setup_count_args(subparsers, parent)
     parser_estimate = setup_estimate_args(subparsers, parent)
     command_to_parser = {
         'ref': parser_ref,
         'align': parser_align,
+        'consensus': parser_consensus,
         'count': parser_count,
         'estimate': parser_estimate,
     }
