@@ -1,33 +1,36 @@
+import multiprocessing
 import os
 import shutil
 import tempfile
 from collections import Counter
 from functools import partial
+from typing import List, Optional, Set, Tuple, Union
 
 import ngs_tools as ngs
 import numpy as np
 import pandas as pd
 import pysam
+from typing_extensions import Literal
 
 from .. import config, utils
 from ..logging import logger
 
-CONVERSION_CSV_COLUMNS = ['read_id', 'index', 'contig', 'genome_i', 'original', 'converted', 'quality']
+CONVERSION_CSV_COLUMNS = ['read_id', 'index', 'contig', 'genome_i', 'conversion', 'quality']
 ALIGNMENT_COLUMNS = [
     'read_id', 'index', 'barcode', 'umi', 'GX', 'A', 'C', 'G', 'T', 'velocity', 'transcriptome', 'score'
 ]
 
 
-def read_alignments(alignments_path, *args, **kwargs):
+def read_alignments(alignments_path: str, *args, **kwargs) -> pd.DataFrame:
     """Read alignments CSV as a pandas DataFrame.
 
     Any additional arguments and keyword arguments are passed to `pandas.read_csv`.
 
-    :param alignments_path: path to alignments CSV
-    :type alignments_path: str
+    Args:
+        alignments_path: path to alignments CSV
 
-    :return: conversions dataframe
-    :rtype: pandas.DataFrame
+    Returns:
+        Conversions dataframe
     """
     df = pd.read_csv(
         alignments_path,
@@ -52,16 +55,16 @@ def read_alignments(alignments_path, *args, **kwargs):
     return df
 
 
-def read_conversions(conversions_path, *args, **kwargs):
+def read_conversions(conversions_path: str, *args, **kwargs) -> pd.DataFrame:
     """Read conversions CSV as a pandas DataFrame.
 
     Any additional arguments and keyword arguments are passed to `pandas.read_csv`.
 
-    :param conversions_path: path to conversions CSV
-    :type conversions_path: str
+    Args:
+        conversions_path: Path to conversions CSV
 
-    :return: conversions dataframe
-    :rtype: pandas.DataFrame
+    Returns:
+        Conversions dataframe
     """
     df = pd.read_csv(
         conversions_path,
@@ -70,8 +73,8 @@ def read_conversions(conversions_path, *args, **kwargs):
             'index': np.uint8,
             'contig': 'category',
             'genome_i': np.uint32,
-            'original': 'category',
-            'converted': 'category',
+            'conversion': 'category',
+            'quality': np.uint8
         },
         na_filter=False,
         *args,
@@ -80,22 +83,25 @@ def read_conversions(conversions_path, *args, **kwargs):
     return df
 
 
-def select_alignments(df_alignments):
+def select_alignments(df_alignments: pd.DataFrame) -> Set[Tuple[str, str]]:
     """Select alignments among duplicates. This function performs preliminary
     deduplication and returns a list of tuples (read_id, alignment index) to
     use for coverage calculation and SNP detection.
 
-    :param df_alignments: alignments dataframe
-    :type df_alignments: pandas.DataFrame
+    Args:
+        df_alignments: Alignments dataframe
 
-    :return: set of (read_id, alignment index) that were selected
-    :rtype: set
+    Returns:
+        Set of (read_id, alignment index) that were selected
     """
     df_sorted = df_alignments.sort_values(['transcriptome', 'score'])
     umi = all(df_alignments['umi'] != 'NA')
     if umi:
         df_deduplicated = df_sorted.drop_duplicates(subset=['barcode', 'umi', 'GX'], keep='last')
     else:
+        # TODO: This part can probably be removed because BAM parsing only considers
+        # primary alignments now.
+
         # Drop any multimappers where none map to the transcriptome and
         # are assigned multiple genes
         df_sorted['not_transcriptome'] = ~df_sorted['transcriptome']
@@ -113,70 +119,60 @@ def select_alignments(df_alignments):
 
 
 def parse_read_contig(
-    counter,
-    lock,
-    bam_path,
-    contig,
-    gene_infos=None,
-    transcript_infos=None,
-    strand='forward',
-    umi_tag=None,
-    barcode_tag=None,
-    gene_tag='GX',
-    barcodes=None,
-    temp_dir=None,
-    update_every=2000,
-    nasc=False,
-    velocity=True
-):
+    counter: multiprocessing.Value,
+    lock: multiprocessing.Lock,
+    bam_path: str,
+    contig: str,
+    gene_infos: Optional[dict] = None,
+    transcript_infos: Optional[dict] = None,
+    strand: Literal['forward', 'reverse', 'unstranded'] = 'forward',
+    umi_tag: Optional[str] = None,
+    barcode_tag: Optional[str] = None,
+    gene_tag: str = 'GX',
+    barcodes: Optional[List[str]] = None,
+    temp_dir: Optional[str] = None,
+    update_every: int = 2000,
+    nasc: bool = False,
+    velocity: bool = True,
+    strict_exon_overlap: bool = False,
+) -> Tuple[str, str, str]:
     """Parse all reads mapped to a contig, outputing conversion
     information as temporary CSVs. This function is designed to be called as a
     separate process.
 
-    :param counter: counter that keeps track of how many reads have been processed
-    :type counter: multiprocessing.Value
-    :param lock: semaphore for the `counter` so that multiple processes do not
-                 modify it at the same time
-    :type lock: multiprocessing.Lock
-    :param bam_path: path to alignment BAM file
-    :type bam_path: str
-    :param contig: only reads that map to this contig will be processed
-    :type contig: str
-    :param gene_infos: dictionary containing gene information, as returned by
-                       `preprocessing.gtf.parse_gtf`, required if `velocity=True`,
-                       defaults to `None`
-    :type gene_infos: dictionary
-    :param transcript_infos: dictionary containing transcript information,
-                             as returned by `preprocessing.gtf.parse_gtf`,
-                             required if `velocity=True`, defaults to `None`
-    :type transcript_infos: dictionary
-    :param strand: strandedness of the sequencing protocol, defaults to `forward`,
-                   may be one of the following: `forward`, `reverse`, `None` (unstranded)
-    :type strand: str, optional
-    :param umi_tag: BAM tag that encodes UMI, if not provided, `NA` is output in the
-                    `umi` column, defaults to `None`
-    :type umi_tag: str, optional
-    :param barcode_tag: BAM tag that encodes cell barcode, if not provided, `NA`
-                        is output in the `barcode` column, defaults to `None`
-    :type barcode_tag: str, optional
-    :param gene_tag: BAM tag that encodes gene assignment, defaults to `GX`
-    :type gene_tag: str, optional
-    :param barcodes: list of barcodes to be considered. All barcodes are considered
-                     if not provided, defaults to `None`
-    :type barcodes: list, optional
-    :param temp_dir: path to temporary directory, defaults to `None`
-    :type temp_dir: str, optional
-    :param update_every: update the counter every this many reads, defaults to `5000`
-    :type update_every: int, optional
-    :param nasc: flag to change behavior to match NASC-seq pipeline, defaults to `False`
-    :type nasc: bool, optional
-    :param velocity: whether or not to assign a velocity type to each read,
-                     defaults to `True`
-    :type velocity: bool, optional
+    Args:
+        counter: Counter that keeps track of how many reads have been processed
+        lock: Semaphore for the `counter` so that multiple processes do not
+            modify it at the same time
+        bam_path: Path to alignment BAM file
+        contig: Only reads that map to this contig will be processed
+        gene_infos: Dictionary containing gene information, as returned by
+            `preprocessing.gtf.parse_gtf`, required if `velocity=True`
+        transcript_infos: Dictionary containing transcript information,
+            as returned by `preprocessing.gtf.parse_gtf`, required if `velocity=True`
+        strand: Strandedness of the sequencing protocol, defaults to `forward`,
+            may be one of the following: `forward`, `reverse`, `unstranded`
+        umi_tag: BAM tag that encodes UMI, if not provided, `NA` is output in the
+            `umi` column
+        barcode_tag: BAM tag that encodes cell barcode, if not provided, `NA`
+            is output in the `barcode` column
+        gene_tag: BAM tag that encodes gene assignment
+        barcodes: List of barcodes to be considered. All barcodes are considered
+            if not provided
+        temp_dir: Path to temporary directory
+        update_every: Update the counter every this many reads, defaults to `5000`
+        nasc: Flag to change behavior to match NASC-seq pipeline, defaults to `False`
+        velocity: Whether or not to assign a velocity type to each read
+        strict_exon_overlap: Whether to use a stricter algorithm to assin reads
+            as spliced
 
-    :return: (path to conversions, path to conversions index, path to alignments)
-    :rtype: (str, str, str)
+    Returns:
+        (path to conversions, path to conversions index, path to alignments)
     """
+
+    def skip_alignment(read, tags):
+        return read.is_secondary or read.is_unmapped or read.is_duplicate or any(not read.has_tag(tag) for tag in tags)
+
     # Sort genes by segment positions
     if velocity:
         gene_order = sorted(gene_infos.keys(), key=lambda gene: tuple(gene_infos[gene]['segment']))
@@ -222,9 +218,9 @@ def parse_read_contig(
                 if not any_exon_overlap and not any_intron_overlap:
                     continue
 
-            if any_exon_only and not any_intron_overlap:
+            if any_exon_only and (not strict_exon_overlap or not any_intron_overlap):
                 assigned_gene, assigned_type = gene, 'spliced'
-            elif not any_exon_only and any_intron_overlap:
+            elif any_intron_overlap and (not strict_exon_overlap or not any_exon_only):
                 assigned_gene, assigned_type = gene, 'unspliced'
             else:
                 assigned_gene, assigned_type = gene, 'ambiguous'
@@ -263,10 +259,10 @@ def parse_read_contig(
 
             # Update every some interval. Updating every loop is very slow.
             n += 1
-            if n % update_every == 0:
-                lock.acquire()
-                counter.value += update_every
-                lock.release()
+            if n == update_every:
+                with lock:
+                    counter.value += update_every
+                n = 0
 
                 # Update gene order (remove genes we no longer have to consider)
                 # We take advantage of the fact that gene_order contains genes sorted
@@ -287,13 +283,12 @@ def parse_read_contig(
                     if to_remove > 0:
                         gene_order = gene_order[to_remove:]
 
-            # Skip reads that do not contain these tags
-            if any(not read.has_tag(tag) for tag in required_tags) or read.is_duplicate or read.is_unmapped:
+            # Skip reads
+            if skip_alignment(read, required_tags):
                 continue
 
             barcode = read.get_tag(barcode_tag) if barcode_tag else 'NA'
-            # If barcodes are provided, skip any barcodes not in the list.
-            if barcodes and barcode not in barcodes:
+            if barcode == '-' or (barcodes and barcode not in barcodes):
                 continue
 
             read_id = read.query_name
@@ -317,7 +312,7 @@ def parse_read_contig(
                     continue
 
                 mate = paired[key]
-                mate_sequence = mate.seq.upper()
+                mate_sequence = mate.query_sequence.upper()
                 mate_qualities = mate.query_qualities
                 mate_reference = mate.get_reference_sequence().upper()
                 mate_reference_positions = mate.get_reference_positions()
@@ -359,7 +354,7 @@ def parse_read_contig(
             umi = read.get_tag(umi_tag) if umi_tag else 'NA'
             gx_assigned = read.has_tag(gene_tag)
             gx = read.get_tag(gene_tag) if gx_assigned else ''
-            sequence = read.seq.upper()
+            sequence = read.query_sequence.upper()
             qualities = read.query_qualities
             reference = read.get_reference_sequence().upper()
             counts += Counter(reference)
@@ -368,7 +363,18 @@ def parse_read_contig(
             assignment = 'unassigned'
             if velocity:
                 read_strand = None
-                if strand == 'forward':
+                if read.is_paired:
+                    if read.is_read1:  # R1 is mapped after R2
+                        if strand == 'forward':
+                            read_strand = '+' if read.is_reverse else '-'
+                        elif strand == 'reverse':
+                            read_strand = '-' if read.is_reverse else '+'
+                    else:  # R1 is mapped before R2
+                        if strand == 'forward':
+                            read_strand = '-' if read.is_reverse else '+'
+                        elif strand == 'reverse':
+                            read_strand = '+' if read.is_reverse else '-'
+                elif strand == 'forward':
                     read_strand = '-' if read.is_reverse else '+'
                 elif strand == 'reverse':
                     read_strand = '+' if read.is_reverse else '-'
@@ -401,7 +407,7 @@ def parse_read_contig(
 
                     conversions_out.write(
                         f'{read_id},{alignment_index},{contig},{genome_i},'
-                        f'{genome_base},{read_base},{quality}\n'
+                        f'{genome_base}{read_base},{quality}\n'
                     )
 
             # Flush cached conversions
@@ -409,7 +415,7 @@ def parse_read_contig(
                 n_lines += 1
                 conversions_out.write(
                     f'{read_id},{alignment_index},{contig},{genome_i},'
-                    f'{genome_base},{read_base},{quality}\n'
+                    f'{genome_base}{read_base},{quality}\n'
                 )
 
             # Add to index if lines were written
@@ -422,34 +428,61 @@ def parse_read_contig(
                 f'{assignment},{gx_assigned},{alignment_score}\n'
             )
 
-    lock.acquire()
-    counter.value += n % update_every
-    lock.release()
+    with lock:
+        counter.value += n
 
     # Save index
     index_path = utils.write_pickle(index, index_path, protocol=4)
 
+    if gene_infos:
+        del gene_infos
+    if transcript_infos:
+        del transcript_infos
+    if barcodes:
+        del barcodes
+
     return conversions_path, index_path, alignments_path
 
 
-def check_bam_tags_exist(bam_path, tags, n_reads=100000, n_threads=8):
+def get_tags_from_bam(bam_path: str, n_reads: int = 100000, n_threads: int = 8) -> Set[str]:
+    """Utility function to retrieve all read tags present in a BAM.
+
+    Args:
+        bam_path: Path to BAM
+        n_reads: Number of reads to consider
+        n_threads: Number of threads
+
+    Returns:
+        Set of all tags found
+    """
+    tags = []
+    with pysam.AlignmentFile(bam_path, 'rb', threads=n_threads) as bam:
+        for i, read in enumerate(bam.fetch(until_eof=True)):
+            for (tag, _) in read.get_tags():
+                tags.append(tag)
+            if i + 1 >= n_reads:
+                break
+    return set(tags)
+
+
+def check_bam_tags_exist(bam_path: str,
+                         tags: List[str],
+                         n_reads: int = 100000,
+                         n_threads: int = 8) -> Tuple[bool, List[str]]:
     """Utility function to check if BAM tags exists in a BAM within the first
     `n_reads` reads.
 
-    :param bam_path: path to BAM
-    :type bam_path: str
-    :param tags: tags to check for
-    :type tags: list
-    :param n_reads: number of reads to consider, defaults to `100000`
-    :type n_reads: int, optional
-    :param n_threads: number of threads, defaults to `8`
-    :type n_threads: int, optional
+    Args:
+        bam_path: Path to BAM
+        tags: Tags to check for
+        n_reads: Number of reads to consider
+        n_threads: Number of threads
 
-    :return: (whether all tags were found, list of not found tags)
-    :rtype: (bool, list)
+    Returns:
+        (whether all tags were found, list of not found tags)
     """
     tags_found = {tag: False for tag in tags}
-    with pysam.AlignmentFile(bam_path, 'rb') as bam:
+    with pysam.AlignmentFile(bam_path, 'rb', threads=n_threads) as bam:
         for i, read in enumerate(bam.fetch()):
             for tag in tags:
                 if read.has_tag(tag):
@@ -462,20 +495,17 @@ def check_bam_tags_exist(bam_path, tags, n_reads=100000, n_threads=8):
     return False, [tag for tag in tags_found if not tags_found[tag]]
 
 
-def check_bam_is_paired(bam_path, n_reads=100000, n_threads=8):
+def check_bam_is_paired(bam_path: str, n_reads: int = 100000, n_threads: int = 8) -> bool:
     """Utility function to check if BAM has paired reads.
 
-    :param bam_path: path to BAM
-    :type bam_path: str
-    :param n_reads: number of reads to consider, defaults to `100000`
-    :type n_reads: int, optional
-    :param n_threads: number of threads, defaults to `8`
-    :type n_threads: int, optional
+        bam_path: Path to BAM
+        n_reads: Number of reads to consider
+        n_threads: Number of threads
 
-    :return: whether paired reads were detected
-    :rtype: bool
+    Returns:
+        Whether paired reads were detected
     """
-    with pysam.AlignmentFile(bam_path, 'rb') as bam:
+    with pysam.AlignmentFile(bam_path, 'rb', threads=n_threads) as bam:
         for i, read in enumerate(bam.fetch()):
             if read.is_paired:
                 return True
@@ -485,91 +515,109 @@ def check_bam_is_paired(bam_path, n_reads=100000, n_threads=8):
     return False
 
 
-def parse_all_reads(
-    bam_path,
-    conversions_path,
-    alignments_path,
-    index_path,
-    gene_infos,
-    transcript_infos,
-    strand='forward',
-    umi_tag=None,
-    barcode_tag=None,
-    gene_tag='GX',
-    barcodes=None,
-    n_threads=8,
-    temp_dir=None,
-    nasc=False,
-    control=False,
-    velocity=True
-):
-    """Parse all reads in a BAM and extract conversion, content and alignment
-    information as CSVs.
+def check_bam_contains_secondary(bam_path: str, n_reads: int = 100000, n_threads: int = 8) -> bool:
+    """Check whether BAM contains secondary alignments.
 
-    :param bam_path: path to alignment BAM file
-    :type bam_path: str
-    :param conversions_path: path to output information about reads that have conversions
-    :type conversions_path: str
-    :param alignments_path: path to alignments information about reads
-    :type alignments_path: str
-    :param index_path: path to conversions index
-    :type index_path: str
-    :param no_index_path: path to no conversions index
-    :type no_index_path: str
-    :param gene_infos: dictionary containing gene information, as returned by
-                       `ngs.gtf.genes_and_transcripts_from_gtf`
-    :type gene_infos: dictionary
-    :param transcript_infos: dictionary containing transcript information,
-                             as returned by `ngs.gtf.genes_and_transcripts_from_gtf`
-    :type transcript_infos: dictionary
-    :param strand: strandedness of the sequencing protocol, defaults to `forward`,
-                   may be one of the following: `forward`, `reverse`, `unstranded`
-    :type strand: str, optional
-    :param umi_tag: BAM tag that encodes UMI, if not provided, `NA` is output in the
-                    `umi` column, defaults to `None`
-    :type umi_tag: str, optional
-    :param barcode_tag: BAM tag that encodes cell barcode, if not provided, `NA`
-                        is output in the `barcode` column, defaults to `None`
-    :type barcode_tag: str, optional
-    :param gene_tag: BAM tag that encodes gene assignment, defaults to `GX`
-    :type gene_tag: str, optional
-    :param barcodes: list of barcodes to be considered. All barcodes are considered
-                     if not provided, defaults to `None`
-    :type barcodes: list, optional
-    :param n_threads: number of threads, defaults to `8`
-    :type n_threads: int, optional
-    :param temp_dir: path to temporary directory, defaults to `None`
-    :type temp_dir: str, optional
-    :param nasc: flag to change behavior to match NASC-seq pipeline, defaults to `False`
-    :type nasc: bool, optional
-    :param velocity: whether or not to assign a velocity type to each read,
-                     defaults to `True`
-    :type velocity: bool, optional
+        bam_path: Path to BAM
+        n_reads: Number of reads to consider
+        n_threads: Number of threads
 
-    :return: (path to conversions, path to alignments, path to conversions index)
-    :rtype: (str, str, str)
+    Returns:
+        Whether secondary alignments were detected
     """
-    logger.debug('Checking if BAM has required tags')
-    tags = config.BAM_REQUIRED_TAGS.copy()
-    if barcode_tag:
-        tags.append(barcode_tag)
-    if umi_tag:
-        tags.append(umi_tag)
-    if gene_tag:
-        tags.append(gene_tag)
-    all_exists, not_found = check_bam_tags_exist(bam_path, tags, config.BAM_PEEK_READS, n_threads=n_threads)
-    if not all_exists:
-        raise Exception(
-            f'First {config.BAM_PEEK_READS} reads in the BAM do not contain the following required tags: '
-            f'{", ".join(not_found)}. '
-        )
+    with pysam.AlignmentFile(bam_path, 'rb', threads=n_threads) as bam:
+        for i, read in enumerate(bam.fetch()):
+            if read.is_secondary:
+                return True
 
-    contig_genes = {}
-    contig_transcripts = {}
-    for gene_id, gene_info in gene_infos.items():
-        contig_genes.setdefault(gene_info['chromosome'], []).append(gene_id)
-        contig_transcripts.setdefault(gene_info['chromosome'], []).extend(gene_info['transcripts'])
+            if i + 1 >= n_reads:
+                break
+    return False
 
+
+def check_bam_contains_unmapped(bam_path: str) -> bool:
+    """Check whether BAM contains unmapped reads.
+
+        bam_path: Path to BAM
+
+    Returns:
+        Whether unmapped reads were detected
+    """
+    with pysam.AlignmentFile(bam_path, 'rb') as bam:
+        for index in bam.get_index_statistics():
+            if index.unmapped > 0:
+                return True
+    return False
+
+
+def check_bam_contains_duplicate(bam_path, n_reads=100000, n_threads=8) -> bool:
+    """Check whether BAM contains duplicates.
+
+        bam_path: Path to BAM
+        n_reads: Number of reads to consider
+        n_threads: Number of threads
+
+    Returns:
+        Whether duplicates were detected
+    """
+    with pysam.AlignmentFile(bam_path, 'rb', threads=n_threads) as bam:
+        for i, read in enumerate(bam.fetch()):
+            if read.is_duplicate:
+                return True
+
+            if i + 1 >= n_reads:
+                break
+    return False
+
+
+def sort_and_index_bam(bam_path: str, out_path: str, n_threads: int = 8, temp_dir: Optional[str] = None) -> str:
+    """Sort and index BAM.
+
+    If the BAM is already sorted, the sorting step is skipped.
+
+    Args:
+        bam_path: Path to alignment BAM file to sort
+        out_path: Path to output sorted BAM
+        n_threads: Number of threads
+        temp_dir: Path to temporary directory
+
+    Returns:
+        Path to sorted and indexed BAM
+    """
+    bam_sorted = False
+    with pysam.AlignmentFile(bam_path, 'rb') as f:
+        if f.header.get('HD', {}).get('SO') == 'coordinate':
+            bam_sorted = True
+
+    if not bam_sorted:
+        logger.info(f'Sorting {bam_path} with samtools to {out_path}')
+        args = ['-o', out_path, '-@', str(n_threads)]
+        if temp_dir:
+            args.extend(['-T', temp_dir])
+        pysam.sort(bam_path, *args)
+        bam_path = out_path
+
+    # Check if BAM index exists and create one if it doesn't.
+    bai_path = f'{bam_path}.bai'
+    if not utils.all_exists(bai_path) or not bam_sorted:
+        logger.info(f'Indexing {bam_path} with samtools to {bai_path}')
+        pysam.index(bam_path, bai_path, '-@', str(n_threads))
+
+    return bam_path
+
+
+def split_bam(bam_path: str, n: int, n_threads: int = 8, temp_dir: Optional[str] = None) -> List[Tuple[str, int]]:
+    """Split BAM into n parts.
+
+    Args:
+        bam_path: Path to alignment BAM file
+        n: Number of splits
+        n_threads: Number of threads
+        temp_dir: Path to temporary directory
+
+    Returns:
+        List of tuples containing (split BAM path, number of reads)
+    """
     # Filter for properly paired reads if BAM contains paired reads
     if check_bam_is_paired(bam_path, config.BAM_PEEK_READS, n_threads=n_threads):
         logger.debug('BAM contains paired reads. Filtering out non-properly paired reads')
@@ -582,25 +630,89 @@ def parse_all_reads(
         )
         pysam.index(bam_path, f'{bam_path}.bai', '-@', str(n_threads))
 
-    if n_threads > 1:
-        n_splits = min(n_threads * 8, utils.get_file_descriptor_limit() - 10)
-        logger.debug(f'Splitting BAM into {n_splits} parts')
-        splits = ngs.bam.split_bam(
+    logger.debug(f'Splitting BAM into {n} parts')
+    splits = list(
+        ngs.bam.split_bam(
             bam_path,
             utils.mkstemp(temp_dir, delete=True),
-            n=n_splits,
+            n=n,
             n_threads=n_threads,
             show_progress=True,
-        )
-        for path, _ in splits.values():
-            pysam.index(path, f'{path}.bai', '-@', str(n_threads))
+        ).values()
+    )
+    for path, _ in splits:
+        pysam.index(path, f'{path}.bai', '-@', str(n_threads))
+    return splits
 
-    else:
-        splits = {'0': (bam_path, ngs.bam.count_bam(bam_path))}
+
+def parse_all_reads(
+    bam_path: str,
+    conversions_path: str,
+    alignments_path: str,
+    index_path: str,
+    gene_infos: dict,
+    transcript_infos: dict,
+    strand: Literal['forward', 'reverse', 'unstranded'] = 'forward',
+    umi_tag: Optional[str] = None,
+    barcode_tag: Optional[str] = None,
+    gene_tag: str = 'GX',
+    barcodes: Optional[List[str]] = None,
+    n_threads: int = 8,
+    temp_dir: Optional[str] = None,
+    nasc: bool = False,
+    control: bool = False,
+    velocity: bool = True,
+    strict_exon_overlap: bool = False,
+    return_splits: bool = False,
+) -> Union[Tuple[str, str, str], Tuple[str, str, str, List[Tuple[str, int]]]]:
+    """Parse all reads in a BAM and extract conversion, content and alignment
+    information as CSVs.
+
+        bam_path: Path to alignment BAM file
+        conversions_path: Path to output information about reads that have conversions
+        alignments_path: Path to alignments information about reads
+        index_path: Path to conversions index
+        no_index_path: Path to no conversions index
+        gene_infos: Dictionary containing gene information, as returned by
+            `ngs.gtf.genes_and_transcripts_from_gtf`
+        transcript_infos: Dictionary containing transcript information,
+            as returned by `ngs.gtf.genes_and_transcripts_from_gtf`
+        strand: Strandedness of the sequencing protocol, defaults to `forward`,
+            may be one of the following: `forward`, `reverse`, `unstranded`
+        umi_tag: BAM tag that encodes UMI, if not provided, `NA` is output in the
+            `umi` column
+        barcode_tag: BAM tag that encodes cell barcode, if not provided, `NA`
+            is output in the `barcode` column
+        gene_tag: BAM tag that encodes gene assignment, defaults to `GX`
+        barcodes: List of barcodes to be considered. All barcodes are considered
+            if not provided
+        n_threads: Number of threads
+        temp_dir: Path to temporary directory
+        nasc: Flag to change behavior to match NASC-seq pipeline
+        velocity: Whether or not to assign a velocity type to each read
+        strict_exon_overlap: Whether to use a stricter algorithm to assign reads as spliced
+        return_splits: return BAM splits for later reuse
+
+    Returns:
+        (path to conversions, path to alignments, path to conversions index)
+            If `return_splits` is True, then there is an additional return value, which
+            is a list of tuples containing split BAM paths and number of reads
+            in each BAM.
+    """
+    contig_genes = {}
+    contig_transcripts = {}
+    for gene_id, gene_info in gene_infos.items():
+        contig_genes.setdefault(gene_info['chromosome'], []).append(gene_id)
+        contig_transcripts.setdefault(gene_info['chromosome'], []).extend(gene_info['transcripts'])
+
+    splits = split_bam(
+        bam_path, min(n_threads * 8,
+                      utils.get_file_descriptor_limit() - 10), n_threads=n_threads, temp_dir=temp_dir
+    ) if n_threads > 1 else [(bam_path, ngs.bam.count_bam(bam_path))]
 
     # Add argument per contig
     args = []
-    for path, _ in splits.values():
+    for path, _ in splits:
         with pysam.AlignmentFile(path, 'rb') as f:
             for index in f.get_index_statistics():
                 if index.total > 0:
@@ -629,12 +741,13 @@ def parse_all_reads(
             temp_dir=tempfile.mkdtemp(dir=temp_dir),
             nasc=nasc,
             velocity=velocity,
+            strict_exon_overlap=strict_exon_overlap,
         ), args
     )
     pool.close()
 
     # Display progres bar
-    utils.display_progress_with_counter(counter, sum(value[1] for value in splits.values()), async_result)
+    utils.display_progress_with_counter(counter, sum(value[1] for value in splits), async_result)
     pool.join()
 
     # Combine csvs
@@ -663,4 +776,5 @@ def parse_all_reads(
     logger.debug(f'Writing conversions index to {index_path}')
     index_path = utils.write_pickle(index, index_path)
 
-    return conversions_path, alignments_path, index_path
+    return (conversions_path, alignments_path,
+            index_path) if not return_splits else (conversions_path, alignments_path, index_path, splits)
