@@ -1,10 +1,14 @@
+import multiprocessing
 import re
 import shutil
 import tempfile
 from functools import partial
+from typing import Dict, FrozenSet, List, Optional, Set, Tuple
 
+import ngs_tools as ngs
 import numpy as np
 import pandas as pd
+from typing_extensions import Literal
 
 from .. import config, utils
 from ..logging import logger
@@ -15,8 +19,7 @@ CONVERSIONS_PARSER = re.compile(
     (?P<index>[^,]*),
     (?P<contig>[^,]*),
     (?P<genome_i>[^,]*),
-    (?P<original>[^,]*),
-    (?P<converted>[^,]*),
+    (?P<conversion>[^,]*),
     (?P<quality>[^,]*)\n
     $''', re.VERBOSE
 )
@@ -39,18 +42,18 @@ ALIGNMENTS_PARSER = re.compile(
 )
 
 CONVERSION_IDX = {
-    ('A', 'C'): 0,
-    ('A', 'G'): 1,
-    ('A', 'T'): 2,
-    ('C', 'A'): 3,
-    ('C', 'G'): 4,
-    ('C', 'T'): 5,
-    ('G', 'A'): 6,
-    ('G', 'C'): 7,
-    ('G', 'T'): 8,
-    ('T', 'A'): 9,
-    ('T', 'C'): 10,
-    ('T', 'G'): 11,
+    'AC': 0,
+    'AG': 1,
+    'AT': 2,
+    'CA': 3,
+    'CG': 4,
+    'CT': 5,
+    'GA': 6,
+    'GC': 7,
+    'GT': 8,
+    'TA': 9,
+    'TC': 10,
+    'TG': 11,
 }
 BASE_IDX = {
     'A': 12,
@@ -58,22 +61,26 @@ BASE_IDX = {
     'G': 14,
     'T': 15,
 }
-CONVERSION_COLUMNS = [''.join(pair) for pair in sorted(CONVERSION_IDX.keys())]
+CONVERSION_COMPLEMENT = {
+    conversion: ngs.sequence.complement_sequence(conversion, reverse=False)
+    for conversion in CONVERSION_IDX.keys()
+}
+CONVERSION_COLUMNS = sorted(CONVERSION_IDX.keys())
 BASE_COLUMNS = sorted(BASE_IDX.keys())
 COLUMNS = CONVERSION_COLUMNS + BASE_COLUMNS
 CSV_COLUMNS = ['read_id', 'barcode', 'umi', 'GX'] + COLUMNS + ['velocity', 'transcriptome', 'score']
 
 
-def read_counts(counts_path, *args, **kwargs):
+def read_counts(counts_path: str, *args, **kwargs) -> pd.DataFrame:
     """Read counts CSV as a pandas dataframe.
 
     Any additional arguments and keyword arguments are passed to `pandas.read_csv`.
 
-    :param counts_path: path to CSV
-    :type counts_path: str
+    Args:
+        counts_path: Path to CSV
 
-    :return: counts dataframe
-    :rtype: pandas.DataFrame
+    Returns:
+        Counts dataframe
     """
     dtypes = {
         'read_id': 'string',
@@ -89,18 +96,16 @@ def read_counts(counts_path, *args, **kwargs):
     return pd.read_csv(counts_path, dtype=dtypes, na_filter=False, *args, **kwargs)
 
 
-def complement_counts(df_counts, gene_infos):
+def complement_counts(df_counts: pd.DataFrame, gene_infos: dict) -> pd.DataFrame:
     """Complement the counts in the counts dataframe according to gene strand.
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
-    :param gene_infos: dictionary containing gene information, as returned by
-                       `preprocessing.gtf.parse_gtf`
-    :type gene_infos: dictionary
+    Args:
+        df_counts: Counts dataframe
+        gene_infos: Dictionary containing gene information, as returned by
+            `preprocessing.gtf.parse_gtf`
 
-    :return: counts dataframe with counts complemented for reads mapping to genes
-        on the reverse strand
-    :rtype: pandas.DataFrame
+    Returns:
+        counts dataframe with counts complemented for reads mapping to genes on the reverse strand
     """
     # Extract columns that do not need to be complemented
     other_columns = []
@@ -108,11 +113,11 @@ def complement_counts(df_counts, gene_infos):
         if col in COLUMNS:
             continue
         other_columns.append(col)
-    df_counts['strand'] = df_counts['GX'].map(lambda gx: gene_infos[gx]['strand'])
+    forward_strand = df_counts['GX'].map(lambda gx: gene_infos[gx]['strand']) == '+'
 
     columns = other_columns + COLUMNS
-    df_forward = df_counts[df_counts.strand == '+'][columns]
-    df_reverse = df_counts[df_counts.strand == '-'][columns]
+    df_forward = df_counts[forward_strand][columns]
+    df_reverse = df_counts[~forward_strand][columns]
 
     df_reverse.columns = other_columns + CONVERSION_COLUMNS[::-1] + BASE_COLUMNS[::-1]
     df_reverse = df_reverse[columns]
@@ -120,22 +125,45 @@ def complement_counts(df_counts, gene_infos):
     return pd.concat((df_forward, df_reverse), verify_integrity=True)
 
 
-def drop_multimappers(df_counts, conversions=None):
+def subset_counts(
+    df_counts: pd.DataFrame,
+    key: Literal['total', 'transcriptome', 'spliced', 'unspliced'],
+) -> pd.DataFrame:
+    """Subset the given counts DataFrame to only contain reads of the desired key.
+
+    Args:
+        df_count: Counts dataframe
+        key: Read types to subset
+
+    Returns:s
+        Subset dataframe
+    """
+    if key == 'transcriptome':
+        df_counts = df_counts[df_counts['transcriptome']]
+    if key in ('spliced', 'unspliced'):
+        df_counts = df_counts[df_counts['velocity'] == key]
+    return df_counts
+
+
+def drop_multimappers(df_counts: pd.DataFrame, conversions: Optional[FrozenSet[str]] = None) -> pd.DataFrame:
     """Drop multimappings that have the same read ID where
     * some map to the transcriptome while some do not -- drop non-transcriptome alignments
     * none map to the transcriptome AND aligned to multiple genes -- drop all
     * none map to the transcriptome AND assigned multiple velocity types -- set to ambiguous
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
-    :param conversions: conversions to prioritize, defaults to `None`
-    :type conversions: list, optional
+    TODO: This function can probably be removed because BAM parsing only considers
+    primary alignments now.
 
-    :return: counts dataframe with multimappers appropriately filtered
-    :rtype: pandas.DataFrame
+    Args:
+        df_counts: Counts dataframe
+        conversions: Conversions to prioritize
+
+    Returns:
+        Counts dataframe with multimappers appropriately filtered
     """
     columns = list(df_counts.columns)
-    df_counts['conversion_sum'] = df_counts[conversions or CONVERSION_COLUMNS].sum(axis=1)
+    convs = list(conversions) if conversions is not None else CONVERSION_COLUMNS
+    df_counts['conversion_sum'] = df_counts[convs].sum(axis=1)
     df_sorted = df_counts.sort_values(['transcriptome', 'score', 'conversion_sum']).drop(columns='conversion_sum')
     df_counts.drop(columns='conversion_sum', inplace=True)
     df_sorted['not_transcriptome'] = ~df_sorted['transcriptome']
@@ -172,47 +200,59 @@ def drop_multimappers(df_counts, conversions=None):
     return df_deduplicated[columns]
 
 
-def deduplicate_counts(df_counts, conversions=None):
+def deduplicate_counts(
+    df_counts: pd.DataFrame,
+    conversions: Optional[FrozenSet[str]] = None,
+    use_conversions: bool = True
+) -> pd.DataFrame:
     """Deduplicate counts based on barcode, UMI, and gene.
 
     The order of priority is the following.
-    1. Reads that align to the transcriptome (exon only)
-    2. If `conversions` is provided, reads that have a larger sum of such conversions
+    1. If `use_conversions=True`, reads that have at least one such conversion
+    2. Reads that align to the transcriptome (exon only)
+    3. Reads that have highest alignment score
+    4. If `conversions` is provided, reads that have a larger sum of such conversions
        If `conversions` is not provided, reads that have larger sum of all conversions
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
-    :param conversions: conversions to prioritize, defaults to `None`
-    :type conversions: list, optional
+    Args:
+        df_counts: Counts dataframe
+        conversions: Conversions to prioritize, defaults to `None`
+        use_conversions: Prioritize reads that have conversions first
 
-    :return: deduplicated counts dataframe
-    :rtype: pandas.DataFrame
+    Returns:
+        Deduplicated counts dataframe
     """
-    columns = list(df_counts.columns)
-    df_counts['conversion_sum'] = df_counts[conversions or CONVERSION_COLUMNS].sum(axis=1)
+    convs = list(conversions) if conversions is not None else CONVERSION_COLUMNS
+    df_counts['conversion_sum'] = df_counts[convs].sum(axis=1)
+    # Deduplication priority.
+    sort_order = ['transcriptome', 'score', 'conversion_sum']
+    to_remove = ['conversion_sum']
 
-    # Sort by transcriptome last, longest alignment last, least conversion first
-    df_sorted = df_counts.sort_values(['transcriptome', 'score', 'conversion_sum']).drop(columns='conversion_sum')
-    df_counts.drop(columns='conversion_sum', inplace=True)
+    if use_conversions and conversions is not None:
+        df_counts['has_conversions'] = df_counts[list(conversions)].sum(axis=1) > 0
+        sort_order.insert(0, 'has_conversions')
+        to_remove.append('has_conversions')
 
-    # Always select transcriptome read if there are duplicates
-    df_deduplicated = df_sorted.drop_duplicates(
-        subset=['barcode', 'umi', 'GX'], keep='last'
-    ).sort_values(['barcode', 'umi', 'GX']).reset_index(drop=True)
+    # Sort by has desired conversion(s) last, transcriptome last,
+    # best alignment last, most conversions last
+    df_sorted = df_counts.sort_values(sort_order).drop(columns=to_remove)
 
-    return df_deduplicated[columns]
+    # Restore input dataframe.
+    df_counts.drop(columns=to_remove, inplace=True)
+
+    return df_sorted.drop_duplicates(subset=['barcode', 'umi', 'GX'], keep='last').reset_index(drop=True)
 
 
-def drop_multimappers_part(counter, lock, split_path, out_path):
-    drop_multimappers(read_counts(split_path))[CSV_COLUMNS[1:]].to_csv(out_path, header=False, index=False)
-    lock.acquire()
-    counter.value += 1
-    lock.release()
-    return out_path
-
-
-def deduplicate_counts_part(counter, lock, split_path, out_path, conversions=None):
-    deduplicate_counts(
+def drop_multimappers_part(
+    counter: multiprocessing.Value,
+    lock: multiprocessing.Lock,
+    split_path: str,
+    out_path: str,
+    conversions: Optional[FrozenSet[str]] = None
+) -> str:
+    """Helper function to parallelize :func:`drop_multimappers`.
+    """
+    drop_multimappers(
         read_counts(split_path), conversions=conversions
     )[CSV_COLUMNS[1:]].to_csv(
         out_path, header=False, index=False
@@ -223,15 +263,35 @@ def deduplicate_counts_part(counter, lock, split_path, out_path, conversions=Non
     return out_path
 
 
-def split_counts_by_velocity(df_counts):
+def deduplicate_counts_part(
+    counter: multiprocessing.Value,
+    lock: multiprocessing.Lock,
+    split_path: str,
+    out_path: str,
+    conversions: Optional[FrozenSet[str]],
+    use_conversions: bool = True
+):
+    """Helper function to parallelize :func:`deduplicate_multimappers`.
+    """
+    deduplicate_counts(
+        read_counts(split_path), conversions=conversions, use_conversions=use_conversions
+    )[CSV_COLUMNS[1:]].to_csv(
+        out_path, header=False, index=False
+    )
+    lock.acquire()
+    counter.value += 1
+    lock.release()
+    return out_path
+
+
+def split_counts_by_velocity(df_counts: pd.DataFrame) -> Dict[str, pd.DataFrame]:
     """Split the given counts dataframe by the `velocity` column.
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
+    Args:
+        df_counts: Counts dataframe
 
-    :return: dictionary containing `velocity` column values as keys and the
-             subset dataframe as values
-    :rtype: dictionary
+    Returns:
+        Dictionary containing `velocity` column values as keys and the subset dataframe as values
     """
     dfs = {}
     for velocity, df_part in df_counts.groupby('velocity', sort=False, observed=True):
@@ -241,35 +301,28 @@ def split_counts_by_velocity(df_counts):
 
 
 def count_no_conversions(
-    alignments_path,
-    counter,
-    lock,
-    index,
-    barcodes=None,
-    temp_dir=None,
-    update_every=10000,
-):
+    alignments_path: str,
+    counter: multiprocessing.Value,
+    lock: multiprocessing.Lock,
+    index: List[Tuple[int, int, int]],
+    barcodes: Optional[List[str]] = None,
+    temp_dir: Optional[str] = None,
+    update_every: int = 10000,
+) -> str:
     """Count reads that have no conversion.
 
-    :param alignments_path: alignments CSV path
-    :type alignments_path: str
-    :param counter: counter that keeps track of how many reads have been processed
-    :type counter: multiprocessing.Value
-    :param lock: semaphore for the `counter` so that multiple processes do not
-                 modify it at the same time
-    :type lock: multiprocessing.Lock
-    :param index: index for conversions CSV
-    :type index: list
-    :param barcodes: list of barcodes to be considered. All barcodes are considered
-                     if not provided, defaults to `None`
-    :type barcodes: list, optional
-    :param temp_dir: path to temporary directory, defaults to `None`
-    :type temp_dir: str, optional
-    :param update_every: update the counter every this many reads, defaults to `5000`
-    :type update_every: int, optional
+    Args:
+        alignments_path: Alignments CSV path
+        counter: Counter that keeps track of how many reads have been processed
+        lock: Semaphore for the `counter` so that multiple processes do not
+            modify it at the same time
+        index: Index for conversions CSV
+        barcodes: List of barcodes to be considered. All barcodes are considered if not provided
+        temp_dir: Path to temporary directory
+        update_every: Update the counter every this many reads
 
-    :return: path to temporary counts CSV
-    :rtype: str
+    Returns:
+        Path to temporary counts CSV
     """
     count_path = utils.mkstemp(dir=temp_dir)
     positions = set(tup[2] for tup in index)
@@ -282,10 +335,11 @@ def count_no_conversions(
             if not line:
                 break
             n += 1
-            if n % update_every == 0:
+            if n == update_every:
                 lock.acquire()
                 counter.value += update_every
                 lock.release()
+                n = 0
             if pos in positions:
                 continue
 
@@ -298,74 +352,64 @@ def count_no_conversions(
                 f'{groups["velocity"]},{groups["transcriptome"]},{groups["score"]}\n'
             )
     lock.acquire()
-    counter.value += n % update_every
+    counter.value += n
     lock.release()
 
     return count_path
 
 
 def count_conversions_part(
-    conversions_path,
-    alignments_path,
-    counter,
-    lock,
-    index,
-    barcodes=None,
-    snps=None,
-    quality=27,
-    temp_dir=None,
-    update_every=10000,
-):
+    conversions_path: str,
+    alignments_path: str,
+    counter: multiprocessing.Value,
+    lock: multiprocessing.Lock,
+    index: List[Tuple[int, int, int]],
+    barcodes: Optional[List[str]] = None,
+    snps: Optional[Dict[str, Dict[str, Set[int]]]] = None,
+    quality: int = 27,
+    temp_dir: Optional[str] = None,
+    update_every: int = 10000,
+) -> str:
     """Count the number of conversions of each read per barcode and gene, along with
     the total nucleotide content of the region each read mapped to, also per barcode
     and gene. This function is used exclusively for multiprocessing.
 
-    :param conversions_path: path to conversions CSV
-    :type conversions_path: str
-    :param alignments_path: path to alignments information about reads
-    :type alignments_path: str
-    :param counter: counter that keeps track of how many reads have been processed
-    :type counter: multiprocessing.Value
-    :param lock: semaphore for the `counter` so that multiple processes do not
-                 modify it at the same time
-    :type lock: multiprocessing.Lock
-    :param index: list of (file position, number of lines) tuples to process
-    :type index: list
-    :param barcodes: list of barcodes to be considered. All barcodes are considered
-                     if not provided, defaults to `None`
-    :type barcodes: list, optional
-    :param snps: dictionary of contig as keys and list of genomic positions as
-                 values that indicate SNP locations, defaults to `None`
-    :type snps: dictionary, optional
-    :param quality: only count conversions with PHRED quality greater than this value,
-                    defaults to `27`
-    :type quality: int, optional
-    :param temp_dir: path to temporary directory, defaults to `None`
-    :type temp_dir: str, optional
-    :param update_every: update the counter every this many reads, defaults to `10000`
-    :type update_every: int, optional
+    Args:
+        conversions_path: Path to conversions CSV
+        alignments_path: Path to alignments information about reads
+        counter: Counter that keeps track of how many reads have been processed
+        lock: Semaphore for the `counter` so that multiple processes do not
+            modify it at the same time
+        index: Index for conversions CSV
+        barcodes: List of barcodes to be considered. All barcodes are considered if not provided
+        snps: Dictionary of contig as keys and list of genomic positions as
+            values that indicate SNP locations
+        quality: Only count conversions with PHRED quality greater than this value
+        temp_dir: Path to temporary directory, defaults to `None`
+        update_every: Update the counter every this many reads
 
-    :return: path to temporary counts CSV
-    :rtype: tuple
+    Returns:
+        Path to temporary counts CSV
     """
 
-    def is_snp(g):
+    def is_snp(gx, conversion, contig, genome_i):
         if not snps:
             return False
-        return int(g['genome_i']) in snps.get(g['contig'], set())
+        return genome_i in snps.get(conversion, {}).get(contig, set())
 
     count_path = utils.mkstemp(dir=temp_dir)
 
     n = 0
-    with open(conversions_path, 'r') as f, open(alignments_path) as f_alignments, open(count_path, 'w') as out:
+    with open(conversions_path, 'r') as f, open(alignments_path, 'r') as f_alignments, open(count_path, 'w') as out:
         for pos, n_lines, pos2 in index:
             f.seek(pos)
             f_alignments.seek(pos2)
             n += 1
-            if n % update_every == 0:
+            if n == update_every:
                 lock.acquire()
                 counter.value += update_every
                 lock.release()
+                n = 0
 
             alignment = ALIGNMENTS_PARSER.match(f_alignments.readline()).groupdict()
             if barcodes and alignment['barcode'] not in barcodes:
@@ -373,10 +417,13 @@ def count_conversions_part(
             counts = [0] * (len(CONVERSION_IDX) + len(BASE_IDX))
             for base, i in BASE_IDX.items():
                 counts[i] = alignment[base]
+            gx = alignment['GX']
             for _ in range(n_lines):
                 groups = CONVERSIONS_PARSER.match(f.readline()).groupdict()
-                if int(groups['quality']) > quality and not is_snp(groups):
-                    counts[CONVERSION_IDX[(groups['original'], groups['converted'])]] += 1
+                conversion = groups["conversion"]
+                if int(groups['quality']) > quality and not is_snp(gx, conversion, groups['contig'], int(
+                        groups['genome_i'])):
+                    counts[CONVERSION_IDX[conversion]] += 1
 
             out.write(
                 f'{groups["read_id"]},{alignment["barcode"]},{alignment["umi"]},'
@@ -385,60 +432,51 @@ def count_conversions_part(
             )
 
         lock.acquire()
-        counter.value += n % update_every
+        counter.value += n
         lock.release()
 
     return count_path
 
 
 def count_conversions(
-    conversions_path,
-    alignments_path,
-    index_path,
-    counts_path,
-    gene_infos,
-    barcodes=None,
-    snps=None,
-    quality=27,
-    conversions=None,
-    n_threads=8,
-    temp_dir=None
-):
+    conversions_path: str,
+    alignments_path: str,
+    index_path: str,
+    counts_path: str,
+    gene_infos: dict,
+    barcodes: Optional[List[str]] = None,
+    snps: Optional[Dict[str, Dict[str, Set[int]]]] = None,
+    quality: int = 27,
+    conversions: Optional[FrozenSet[str]] = None,
+    dedup_use_conversions: bool = True,
+    n_threads: int = 8,
+    temp_dir: Optional[str] = None
+) -> str:
     """Count the number of conversions of each read per barcode and gene, along with
     the total nucleotide content of the region each read mapped to, also per barcode.
     When a duplicate UMI for a barcode is observed, the read with the greatest
     number of conversions is selected.
 
-    :param conversions_path: path to conversions CSV
-    :type conversions_path: str
-    :param alignments_path: path to alignments information about reads
-    :type alignments_path: str
-    :param index_path: path to conversions index
-    :type index_path: str
-    :param counts_path: path to write counts CSV
-    :param counts_path: str
-    :param gene_infos: dictionary containing gene information, as returned by
-                       `ngs.gtf.genes_and_transcripts_from_gtf`, defaults to `None`
-    :type gene_infos: dictionary
-    :param barcodes: list of barcodes to be considered. All barcodes are considered
-                     if not provided, defaults to `None`
-    :type barcodes: list, optional
-    :param snps: dictionary of contig as keys and list of genomic positions as
-                 values that indicate SNP locations, defaults to `None`
-    :type snps: dictionary, optional
-    :param conversions: conversions to prioritize when deduplicating only applicable
-                        for UMI technologies, defaults to `None`
-    :type conversions: list, optional
-    :param quality: only count conversions with PHRED quality greater than this value,
-                    defaults to `27`
-    :type quality: int, optional
-    :param n_threads: number of threads, defaults to `8`
-    :type n_threads: int, optional
-    :param temp_dir: path to temporary directory, defaults to `None`
-    :type temp_dir: str, optional
+    Args:
+        conversions_path: Path to conversions CSV
+        alignments_path: Path to alignments information about reads
+        index_path: Path to conversions index
+        counts_path: Path to write counts CSV
+        gene_infos: Dictionary containing gene information, as returned by
+            `ngs.gtf.genes_and_transcripts_from_gtf`
+        barcodes: List of barcodes to be considered. All barcodes are considered if not provided
+        snps: Dictionary of contig as keys and list of genomic positions as
+            values that indicate SNP locations
+        conversions: Conversions to prioritize when deduplicating only applicable
+            for UMI technologies
+        dedup_use_conversions: Prioritize reads that have at least one conversion
+            when deduplicating
+        quality: Only count conversions with PHRED quality greater than this value
+        n_threads: Number of threads
+        temp_dir: Path to temporary directory
 
-    :return: path to counts CSV
-    :rtype: str
+    Returns:
+        Path to counts CSV
     """
     # Load index
     logger.debug(f'Loading index {index_path}')
@@ -502,7 +540,8 @@ def count_conversions(
     logger.debug(f'Loading combined counts from {combined_path}')
     df_counts = complement_counts(read_counts(combined_path), gene_infos)
     umi = all(df_counts['umi'] != 'NA')
-    barcode_counts = dict(df_counts['barcode'].value_counts(sort=False))
+    barcode_groupby = df_counts.groupby('barcode', sort=False, observed=True)
+    barcode_counts = dict(barcode_groupby.size())
     split_paths = []
     current_split_path = None
     current_split_f = None
@@ -510,8 +549,7 @@ def count_conversions(
     # Split barcodes into approximately `config.COUNTS_SPLIT_THRESHOLD` bins.
     # Note that a single barcode may have more than this many reads.
     try:
-        for barcode in barcode_counts.keys():
-            df_counts_barcode = df_counts[df_counts['barcode'] == barcode]
+        for barcode, df_counts_barcode in barcode_groupby:
             # Make its own split
             if barcode_counts[barcode] > config.COUNTS_SPLIT_THRESHOLD:
                 split_path = utils.mkstemp(dir=temp_dir)
@@ -551,7 +589,8 @@ def count_conversions(
             counter,
             lock,
             conversions=conversions,
-        ) if umi else partial(drop_multimappers_part, counter, lock), paths
+            use_conversions=dedup_use_conversions,
+        ) if umi else partial(drop_multimappers_part, counter, lock, conversions=conversions), paths
     )
     pool.close()
 
