@@ -1,3 +1,4 @@
+import concurrent
 import functools
 import multiprocessing
 import os
@@ -6,12 +7,13 @@ import sys
 import time
 from concurrent.futures import as_completed
 from contextlib import contextmanager
+from typing import Any, Dict, FrozenSet, Iterable, List, Optional, Tuple, Union
 
 import anndata
 import ngs_tools as ngs
 import numpy as np
-import psutil
 import pandas as pd
+import psutil
 from scipy import sparse
 
 from . import config
@@ -26,7 +28,7 @@ flatten_dict_values = ngs.utils.flatten_dict_values
 mkstemp = ngs.utils.mkstemp
 all_exists = ngs.utils.all_exists
 flatten_dictionary = ngs.utils.flatten_dictionary
-flatten_list = ngs.utils.flatten_list
+flatten_iter = ngs.utils.flatten_iter
 merge_dictionaries = ngs.utils.merge_dictionaries
 write_pickle = ngs.utils.write_pickle
 read_pickle = ngs.utils.read_pickle
@@ -66,12 +68,12 @@ class suppress_stdout_stderr:
             os.close(fd)
 
 
-def get_STAR_binary_path():
+def get_STAR_binary_path() -> str:
     """Get the path to the platform-dependent STAR binary included with
     the installation.
 
-    :return: path to the binary
-    :rtype: str
+    Returns:
+        Path to the binary
     """
     bin_filename = 'STAR.exe' if config.PLATFORM == 'windows' else 'STAR'
     path = os.path.join(config.BINS_DIR, config.PLATFORM, 'STAR', bin_filename)
@@ -80,31 +82,30 @@ def get_STAR_binary_path():
     return path
 
 
-def get_STAR_version():
+def get_STAR_version() -> str:
     """Get the provided STAR version.
 
-    :return: version string
-    :rtype: str
+    Returns:
+        Version string
     """
     p, stdout, stderr = run_executable([get_STAR_binary_path(), '--version'], quiet=True, returncode=1)
     version = stdout.strip()
     return version
 
 
-def combine_arguments(args, additional):
+def combine_arguments(args: Dict[str, Any], additional: Dict[str, Any]) -> Dict[str, Any]:
     """Combine two dictionaries representing command-line arguments.
 
     Any duplicate keys will be merged according to the following procedure:
     1. If the value in both dictionaries are lists, the two lists are combined.
     2. Otherwise, the value in the first dictionary is OVERWRITTEN.
 
-    :param args: original command-line arguments
-    :type args: dictionary
-    :param additional: additional command-line arguments
-    :type additional: dictionary
+    Args:
+        args: Original command-line arguments
+        additional: Additional command-line arguments
 
-    :return: combined command-line arguments
-    :rtype: dictionary
+    Returns:
+        Combined command-line arguments
     """
     new_args = args.copy()
 
@@ -120,14 +121,14 @@ def combine_arguments(args, additional):
     return new_args
 
 
-def arguments_to_list(args):
+def arguments_to_list(args: Dict[str, Any]) -> List[Any]:
     """Convert a dictionary of command-line arguments to a list.
 
-    :param args: command-line arguments
-    :type args: dictionary
+    Args:
+        args: Command-line arguments
 
-    :return: list of command-line arguments
-    :rtype: list
+    Returns:
+        List of command-line arguments
     """
     arguments = []
     for key, value in args.items():
@@ -139,12 +140,12 @@ def arguments_to_list(args):
     return arguments
 
 
-def get_file_descriptor_limit():
+def get_file_descriptor_limit() -> int:
     """Get the current value for the maximum number of open file descriptors
     in a platform-dependent way.
 
-    :return: the current value of the maximum number of open file descriptors.
-    :rtype: int
+    Returns:
+        The current value of the maximum number of open file descriptors.
     """
     if config.PLATFORM == 'windows':
         import win32file
@@ -154,7 +155,7 @@ def get_file_descriptor_limit():
         return resource.getrlimit(resource.RLIMIT_NOFILE)[0]
 
 
-def get_max_file_descriptor_limit():
+def get_max_file_descriptor_limit() -> int:
     """Get the maximum allowed value for the maximum number of open file
     descriptors.
 
@@ -165,8 +166,8 @@ def get_max_file_descriptor_limit():
 
     Similarly, on MacOS, we return a hardcoded 10240.
 
-    :return: maximum allowed value for the maximum number of open file descriptors
-    :rtype: int
+    Returns:
+        Maximum allowed value for the maximum number of open file descriptors
     """
     if config.PLATFORM == 'windows':
         return 8192
@@ -178,16 +179,16 @@ def get_max_file_descriptor_limit():
 
 
 @contextmanager
-def increase_file_descriptor_limit(limit):
+def increase_file_descriptor_limit(limit: int):
     """Context manager that can be used to temporarily increase the maximum
     number of open file descriptors for the current process. The original
     value is restored when execution exits this function.
 
     This is required when running STAR with many threads.
 
-    :param limit: maximum number of open file descriptors will be increased to
-                  this value for the duration of the context
-    :type limit: int
+    Args:
+        limit: Maximum number of open file descriptors will be increased to
+            this value for the duration of the context
     """
     old = None
     if config.PLATFORM == 'windows':
@@ -210,23 +211,23 @@ def increase_file_descriptor_limit(limit):
                 resource.setrlimit(resource.RLIMIT_NOFILE, old)
 
 
-def get_available_memory():
+def get_available_memory() -> int:
     """Get total amount of available memory (total memory - used memory) in bytes.
 
-    :return: available memory in bytes
-    :rtype: int
+    Returns:
+        Available memory in bytes
     """
     return psutil.virtual_memory().available
 
 
-def make_pool_with_counter(n_threads):
+def make_pool_with_counter(n_threads: int) -> Tuple[multiprocessing.Pool, multiprocessing.Value, multiprocessing.Lock]:
     """Create a new Process pool with a shared progress counter.
 
-    :param n_threads: number of processes
-    :type n_threads: int
+    Args:
+        n_threads: Number of processes
 
-    :return: (Process pool, progress counter, lock)
-    :rtype: (multiprocessing.Pool, multiprocessing.Value, multiprocessing.Lock)
+    Returns:
+        Tuple of (Process pool, progress counter, lock)
     """
     manager = multiprocessing.Manager()
     counter = manager.Value('I', 0)
@@ -235,34 +236,33 @@ def make_pool_with_counter(n_threads):
     return pool, counter, lock
 
 
-def display_progress_with_counter(counter, total, *async_results, desc=None):
+def display_progress_with_counter(
+    counter: multiprocessing.Value, total: int, *async_results, desc: Optional[str] = None
+):
     """Display progress bar for displaying multiprocessing progress.
 
-    :param counter: progress counter
-    :type counter: multiprocessing.Value
-    :param total: maximum number of units of processing
-    :type total: int
-    :param *async_results: multiprocessing results to monitor. These are used to
-                           determine when all processes are done.
-    :type *async_results: multiprocessing.pool.AsyncResult
-    :param desc: progress bar description, defaults to `None`
-    :type desc: str, optional
+    Args:
+        counter: Progress counter
+        total: Maximum number of units of processing
+        *async_results: Multiprocessing results to monitor. These are used to
+            determine when all processes are done.
+        desc: Progress bar description
     """
     with ngs.progress.progress(total=total, unit_scale=True, desc=desc) as pbar:
         previous_progress = 0
         while any(not async_result.ready() for async_result in async_results):
-            time.sleep(0.1)
+            time.sleep(0.01)
             progress = counter.value
             pbar.update(progress - previous_progress)
             pbar.refresh()
             previous_progress = progress
 
 
-def as_completed_with_progress(futures):
+def as_completed_with_progress(futures: Iterable[concurrent.futures.Future]):
     """Wrapper around `concurrent.futures.as_completed` that displays a progress bar.
 
-    :param futures: iterator of `concurrent.futures.Future` objects
-    :type futures: iterable
+    Args:
+        Iterator of `concurrent.futures.Future` objects
     """
     with ngs.progress.progress(total=len(futures)) as pbar:
         for future in as_completed(futures):
@@ -270,20 +270,18 @@ def as_completed_with_progress(futures):
             pbar.update(1)
 
 
-def split_index(index, n=8):
+def split_index(index: List[Tuple[int, int, int]], n: int = 8) -> List[List[Tuple[int, int, int]]]:
     """Split a conversions index, which is a list of tuples (file position,
     number of lines, alignment position), one for each read, into `n`
     approximately equal parts. This function is used to split the conversions
     CSV for multiprocessing.
 
-    :param index: index
-    :type index: list
-    :param n: number of splits, defaults to `8`
-    :type n: int, optional
+    Args:
+        index: index
+        n: Number of splits, defaults to `8`
 
-    :return: list of parts, where each part is a list of
-        (file position, number of lines, alignment position) tuples
-    :rtype: list
+    Returns:
+        List of parts
     """
     n_lines = sum(idx[1] for idx in index)
     target = (n_lines // n) + 1  # add one to prevent underflow
@@ -306,27 +304,29 @@ def split_index(index, n=8):
     return parts
 
 
-def downsample_counts(df_counts, proportion=None, count=None, seed=None, group_by=None):
+def downsample_counts(
+    df_counts: pd.DataFrame,
+    proportion: Optional[float] = None,
+    count: Optional[int] = None,
+    seed: Optional[int] = None,
+    group_by: Optional[List[str]] = None
+) -> pd.DataFrame:
     """Downsample the given counts dataframe according to the ``proportion`` or
     ``count`` arguments. One of these two must be provided, but not both. The dataframe
     is assumed to be UMI-deduplicated.
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
-    :param proportion: proportion of reads (UMIs) to keep, defaults to None
-    :type proportion: float, optional
-    :param count: absolute number of reads (UMIs) to keep, defaults to None
-    :type count: int, optional
-    :param seed: random seed, defaults to None
-    :type seed: int, optional
-    :param group_by: Columns in the counts dataframe to use to group entries.
-        When this is provided, UMIs are no longer sampled at random, but instead
-        grouped by this argument, and only groups that have more than ``count`` UMIs
-        are downsampled.
-    :type group_by: list, optional
+    Args:
+        df_counts: Counts dataframe
+        proportion: Proportion of reads (UMIs) to keep
+        count: Absolute number of reads (UMIs) to keep
+        seed: Random seed
+        group_by: Columns in the counts dataframe to use to group entries.
+            When this is provided, UMIs are no longer sampled at random, but instead
+            grouped by this argument, and only groups that have more than ``count`` UMIs
+            are downsampled.
 
-    :return: downsampled counts dataframe
-    :rtype: pandas.DataFrame
+    Returns:
+        Downsampled counts dataframe
     """
     rng = np.random.default_rng(seed)
     if not group_by:
@@ -347,24 +347,26 @@ def downsample_counts(df_counts, proportion=None, count=None, seed=None, group_b
         return pd.concat(dfs)
 
 
-def counts_to_matrix(df_counts, barcodes, features, barcode_column='barcode', feature_column='GX'):
+def counts_to_matrix(
+    df_counts: pd.DataFrame,
+    barcodes: List[str],
+    features: List[str],
+    barcode_column: str = 'barcode',
+    feature_column: str = 'GX'
+) -> sparse.csr_matrix:
     """Convert a counts dataframe to a sparse counts matrix.
 
     Counts are assumed to be appropriately deduplicated.
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
-    :param barcodes: list of barcodes that will map to the rows
-    :type barcodes: list
-    :param features: list of features (i.e. genes) that will map to the columns
-    :type features: list
-    :param barcode_column: column in counts dataframe to use as barcodes, defaults to `barcode`
-    :type barcode_column: str
-    :param feature_column: column in counts dataframe to use as features, defaults to `GX`
-    :type feature_column: str
+    Args:
+        df_counts: Counts dataframe
+        barcodes: List of barcodes that will map to the rows
+        features: List of features (i.e. genes) that will map to the columns
+        barcode_column: Column in counts dataframe to use as barcodes, defaults to `barcode`
+        feature_column: Column in counts dataframe to use as features, defaults to `GX`
 
-    :return: sparse counts matrix
-    :rtype: scipy.sparse.csrmatrix
+    Returns:
+        Sparse counts matrix
     """
     # Transform to index for fast lookup
     barcode_indices = {barcode: i for i, barcode in enumerate(barcodes)}
@@ -378,34 +380,36 @@ def counts_to_matrix(df_counts, barcodes, features, barcode_column='barcode', fe
     return matrix.tocsr()
 
 
-def split_counts(df_counts, barcodes, features, barcode_column='barcode', feature_column='GX', conversions=['TC']):
+def split_counts(
+    df_counts: pd.DataFrame,
+    barcodes: List[str],
+    features: List[str],
+    barcode_column: str = 'barcode',
+    feature_column: str = 'GX',
+    conversions: FrozenSet[str] = frozenset({'TC'})
+) -> Tuple[sparse.csr_matrix, sparse.csr_matrix]:
     """Split counts dataframe into two count matrices by a column.
 
-    :param df_counts: counts dataframe
-    :type df_counts: pandas.DataFrame
-    :param barcodes: list of barcodes that will map to the rows
-    :type barcodes: list
-    :param features: list of features (i.e. genes) that will map to the columns
-    :type features: list
-    :param barcode_column: column in counts dataframe to use as barcodes, defaults to `barcode`
-    :type barcode_column: str, optional
-    :param feature_column: column in counts dataframe to use as features, defaults to `GX`
-    :type feature_column: str, optional
-    :param conversions: conversion(s) in question, defaults to `['TC']`
-    :type conversions: list, optional
+    Args:
+        df_counts: Counts dataframe
+        barcodes: List of barcodes that will map to the rows
+        features: List of features (i.e. genes) that will map to the columns
+        barcode_column: Column in counts dataframe to use as barcodes
+        feature_column: Column in counts dataframe to use as features
+        conversions: Conversion(s) in question
 
-    :return: (count matrix of `conversion==0`, count matrix of `conversion>0`)
-    :rtype: (scipy.sparse.csrmatrix, scipy.sparse.csrmatrix)
+    Returns:
+        count matrix of `conversion==0`, count matrix of `conversion>0`
     """
     matrix_unlabeled = counts_to_matrix(
-        df_counts[(df_counts[conversions] == 0).all(axis=1)],
+        df_counts[(df_counts[list(conversions)] == 0).all(axis=1)],
         barcodes,
         features,
         barcode_column=barcode_column,
         feature_column=feature_column
     )
     matrix_labeled = counts_to_matrix(
-        df_counts[(df_counts[conversions] > 0).any(axis=1)],
+        df_counts[(df_counts[list(conversions)] > 0).any(axis=1)],
         barcodes,
         features,
         barcode_column=barcode_column,
@@ -414,28 +418,30 @@ def split_counts(df_counts, barcodes, features, barcode_column='barcode', featur
     return matrix_unlabeled, matrix_labeled
 
 
-def split_matrix(matrix, pis, barcodes, features):
+def split_matrix_pi(
+    matrix: Union[np.ndarray, sparse.spmatrix], pis: Dict[Tuple[str, str], float], barcodes: List[str],
+    features: List[str]
+) -> Tuple[sparse.csr_matrix, sparse.csr_matrix, sparse.csr_matrix]:
     """Split the given matrix based on provided fraction of new RNA.
 
-    :param matrix: matrix to split
-    :type matrix: numpy.ndarray or scipy.sparse.spmatrix
-    :param pis: dictionary containing pi estimates
-    :type pis: dictionary
-    :param barcodes: all barcodes
-    :type barcodes: list
-    :param features: all features (i.e. genes)
-    :type features: list
+    Args:
+        matrix: Matrix to split
+        pis: Dictionary containing pi estimates
+        barcodes: All barcodes
+        features: All features (i.e. genes)
 
-    :return: (matrix of pi masks, matrix of unlabeled RNA, matrix of labeled RNA)
-    :rtype: (scipy.sparse.spmatrix, scipy.sparse.spmatrix, scipy.sparse.spmatrix)
+    Returns:
+        matrix of pis, matrix of unlabeled RNA, matrix of labeled RNA
     """
     unlabeled_matrix = sparse.lil_matrix((len(barcodes), len(features)), dtype=np.float32)
     labeled_matrix = sparse.lil_matrix((len(barcodes), len(features)), dtype=np.float32)
-    pi_mask = sparse.lil_matrix((len(barcodes), len(features)), dtype=bool)
+    pi_matrix = sparse.lil_matrix((len(barcodes), len(features)), dtype=np.float32)
     barcode_indices = {barcode: i for i, barcode in enumerate(barcodes)}
     feature_indices = {feature: i for i, feature in enumerate(features)}
 
     for (barcode, gx), pi in pis.items():
+        if barcode not in barcode_indices or gx not in feature_indices:
+            continue
         try:
             pi = float(pi)
         except ValueError:
@@ -444,35 +450,88 @@ def split_matrix(matrix, pis, barcodes, features):
         val = matrix[row, col]
         unlabeled_matrix[row, col] = val * (1 - pi)
         labeled_matrix[row, col] = val * pi
-        pi_mask[row, col] = True
+        pi_matrix[row, col] = pi
 
-    return pi_mask.tocsr(), unlabeled_matrix.tocsr(), labeled_matrix.tocsr()
+    return pi_matrix.tocsr(), unlabeled_matrix.tocsr(), labeled_matrix.tocsr()
 
 
-def results_to_adata(df_counts, conversions=['TC'], gene_infos=None, pis=None):
+def split_matrix_alpha(
+    unlabeled_matrix: Union[np.ndarray, sparse.spmatrix], labeled_matrix: Union[np.ndarray, sparse.spmatrix],
+    alphas: Dict[str, float], barcodes: List[str]
+) -> Tuple[sparse.csr_matrix, sparse.csr_matrix, sparse.csr_matrix]:
+    """Split the given matrix based on provided fraction of new RNA.
+
+    Args:
+        unlabeled_matrix: unlabeled matrix
+        labeled_matrix: Labeled matrix
+        alphas: Dictionary containing alpha estimates
+        barcodes: All barcodes
+        features: All features (i.e. genes)
+
+    Returns:
+        matrix of pis, matrix of unlabeled RNA, matrix of labeled RNA
+    """
+    total_matrix = unlabeled_matrix + labeled_matrix
+    est_unlabeled_matrix = sparse.lil_matrix(unlabeled_matrix.shape, dtype=np.float32)
+    est_labeled_matrix = sparse.lil_matrix(labeled_matrix.shape, dtype=np.float32)
+    barcode_indices = {barcode: i for i, barcode in enumerate(barcodes)}
+
+    module = np
+    if sparse.issparse(total_matrix):
+        module = sparse
+
+    for barcode, alpha in alphas.items():
+        if barcode not in barcode_indices:
+            continue
+        try:
+            alpha = np.clip(float(alpha), 0, 1)
+        except ValueError:
+            continue
+        row = barcode_indices[barcode]
+        l_alpha = labeled_matrix[row] / alpha
+        est_labeled_matrix[row] = module.vstack((l_alpha, total_matrix[row])).min(axis=0)
+        est_unlabeled_matrix[row] = total_matrix[row] - est_labeled_matrix[row]
+
+    return est_unlabeled_matrix.tocsr(), est_labeled_matrix.tocsr()
+
+
+def results_to_adata(
+    df_counts: pd.DataFrame,
+    conversions: FrozenSet[FrozenSet[str]] = frozenset({frozenset({'TC'})}),
+    gene_infos: Optional[dict] = None,
+    pis: Optional[Dict[str, Dict[Tuple[str, ...], Dict[Tuple[str, str], float]]]] = None,
+    alphas: Optional[Dict[str, Dict[Tuple[str, ...], Dict[str, float]]]] = None,
+) -> anndata.AnnData:
     """Compile all results to a single anndata.
 
-    :param df_counts: counts dataframe, with complemented reverse strand bases
-    :type df_counts: pandas.DataFrame
-    :param conversions: conversion(s) in question, defaults to `['TC']`
-    :type conversions: list, optional
-    :param gene_infos: dictionary containing gene information, defaults to `None`
-    :type gene_infos: dict, optional
-    :param pis: dictionary of estimated pis, defaults to `None`
-    :type pis: dict, optional
+    Args:
+        df_counts: Counts dataframe, with complemented reverse strand bases
+        conversions: Conversion(s) in question
+        gene_infos: Dictionary containing gene information. If this is not provided,
+            the function assumes gene names are already in the Counts dataframe.
+        pis: Dictionary of estimated pis
+        alphas: Dictionary of estimated alphas
 
-    :return: anndata containing all results
-    :rtype: anndata.AnnData
+    Returns:
+        Anndata containing all results
     """
+    if pis is not None and alphas is not None:
+        raise Exception('Only one of `pis` or `alphas` may be provided.')
+
     pis = pis or {}
-    gene_infos = gene_infos or {}
-    all_conversions = sorted(flatten_list(conversions))
+    alphas = alphas or {}
+    all_conversions = sorted(flatten_iter(conversions))
     transcriptome_exists = df_counts['transcriptome'].any()
     transcriptome_only = df_counts['transcriptome'].all()
     velocities = df_counts['velocity'].unique()
     barcodes = sorted(df_counts['barcode'].unique())
     features = sorted(df_counts['GX'].unique())
-    names = [gene_infos.get(feature, {}).get('gene_name') for feature in features]
+
+    add_names = gene_infos is not None
+    obs = pd.DataFrame(index=pd.Series(barcodes, name='barcode'))
+    var = pd.DataFrame(index=pd.Series(features, name='gene_id' if add_names else 'gene_name'))
+    if add_names:
+        var['gene_name'] = pd.Categorical([gene_infos.get(feature, {}).get('gene_name') for feature in features])
 
     df_counts_transcriptome = df_counts[df_counts['transcriptome']]
     matrix = counts_to_matrix(df_counts_transcriptome, barcodes, features)
@@ -481,6 +540,7 @@ def results_to_adata(df_counts, conversions=['TC'], gene_infos=None, pis=None):
     # Transcriptome reads
     if transcriptome_exists:
         for convs in conversions:
+            convs = sorted(convs)
             # Ignore reads that have other conversions
             other_convs = list(set(all_conversions) - set(convs))
             join = '_'.join(convs)
@@ -494,17 +554,25 @@ def results_to_adata(df_counts, conversions=['TC'], gene_infos=None, pis=None):
             pi = pis.get('transcriptome', {}).get(tuple(convs))
             if pi is not None:
                 (
-                    _,
+                    layers[f'X_{join}_pi_g'],
                     layers[f'X_n_{join}_est'],
                     layers[f'X_l_{join}_est'],
-                ) = split_matrix(layers[f'X_n_{join}'] + layers[f'X_l_{join}'], pi, barcodes, features)
+                ) = split_matrix_pi(layers[f'X_n_{join}'] + layers[f'X_l_{join}'], pi, barcodes, features)
+            alpha = alphas.get('transcriptome', {}).get(tuple(convs))
+            if alpha is not None:
+                obs[f'X_{join}_alpha'] = obs.index.map(alpha)
+                (
+                    layers[f'X_n_{join}_est'],
+                    layers[f'X_l_{join}_est'],
+                ) = split_matrix_alpha(layers[f'X_n_{join}'], layers[f'X_l_{join}'], alpha, barcodes)
     else:
         logger.warning('No reads were assigned to `transcriptome`')
 
     # Total reads
     if not transcriptome_only:
         layers['total'] = counts_to_matrix(df_counts, barcodes, features)
-        for conv in conversions:
+        for convs in conversions:
+            convs = sorted(convs)
             other_convs = list(set(all_conversions) - set(convs))
             join = '_'.join(convs)
             layers[f'unlabeled_{join}'], layers[f'labeled_{join}'] = split_counts(
@@ -513,10 +581,17 @@ def results_to_adata(df_counts, conversions=['TC'], gene_infos=None, pis=None):
             pi = pis.get('total', {}).get(tuple(convs))
             if pi is not None:
                 (
-                    _,
+                    layers[f'total_{join}_pi_g'],
                     layers[f'unlabeled_{join}_est'],
                     layers[f'labeled_{join}_est'],
-                ) = split_matrix(layers[f'unlabeled_{join}'] + layers[f'labeled_{join}'], pi, barcodes, features)
+                ) = split_matrix_pi(layers[f'unlabeled_{join}'] + layers[f'labeled_{join}'], pi, barcodes, features)
+            alpha = alphas.get('total', {}).get(tuple(convs))
+            if alpha is not None:
+                obs[f'total_{join}_alpha'] = obs.index.map(alpha)
+                (
+                    layers[f'unlabeled_{join}_est'],
+                    layers[f'labeled_{join}_est'],
+                ) = split_matrix_alpha(layers[f'unlabeled_{join}'], layers[f'labeled_{join}'], alpha, barcodes)
 
     # Velocity reads
     for key in velocities:
@@ -527,7 +602,8 @@ def results_to_adata(df_counts, conversions=['TC'], gene_infos=None, pis=None):
         if key in config.VELOCITY_BLACKLIST:
             continue
 
-        for conv in conversions:
+        for convs in conversions:
+            convs = sorted(convs)
             other_convs = list(set(all_conversions) - set(convs))
             join = '_'.join(convs)
             layers[f'{key[0]}n_{join}'], layers[f'{key[0]}l_{join}'] = split_counts(
@@ -539,18 +615,20 @@ def results_to_adata(df_counts, conversions=['TC'], gene_infos=None, pis=None):
             pi = pis.get(key, {}).get(tuple(convs))
             if pi is not None:
                 (
-                    _,
+                    layers[f'{key}_{join}_pi_g'],
                     layers[f'{key[0]}n_{join}_est'],
                     layers[f'{key[0]}l_{join}_est'],
-                ) = split_matrix(layers[f'{key[0]}n_{join}'] + layers[f'{key[0]}l_{join}'], pi, barcodes, features)
+                ) = split_matrix_pi(layers[f'{key[0]}n_{join}'] + layers[f'{key[0]}l_{join}'], pi, barcodes, features)
+            alpha = alphas.get(key, {}).get(tuple(convs))
+            if alpha is not None:
+                obs[f'{key}_{join}_alpha'] = obs.index.map(alpha)
+                (
+                    layers[f'{key[0]}n_{join}_est'],
+                    layers[f'{key[0]}l_{join}_est'],
+                ) = split_matrix_alpha(layers[f'{key[0]}n_{join}'], layers[f'{key[0]}l_{join}'], alpha, barcodes)
 
     # Construct anndata
-    return anndata.AnnData(
-        X=matrix,
-        obs=pd.DataFrame(index=pd.Series(barcodes, name='barcode')),
-        var=pd.DataFrame(index=pd.Series(features, name='gene_id'), data={'gene_name': pd.Categorical(names)}),
-        layers=layers
-    )
+    return anndata.AnnData(X=matrix, obs=obs, var=var, layers=layers)
 
 
 def patch_mp_connection_bpo_17560():
@@ -604,3 +682,25 @@ def patch_mp_connection_bpo_17560():
 
     Connection._send_bytes = send_bytes
     Connection._recv_bytes = recv_bytes
+
+
+def dict_to_matrix(d: Dict[Tuple[str, str], float], rows: List[str], columns: List[str]) -> sparse.csr_matrix:
+    """Convert a dictionary to a matrix.
+
+    Args:
+        d: Dictionary to convert
+        rows: Row names
+        columns: Column names
+
+    Returns:
+        A sparse matrix
+    """
+    # Transform to index for fast lookup
+    row_indices = {col: i for i, col in enumerate(columns)}
+    column_indices = {row: i for i, row in enumerate(rows)}
+
+    matrix = sparse.lil_matrix((len(rows), len(columns)), dtype=np.float32)
+    for (row, col), value in d.items():
+        matrix[row_indices[row], column_indices[col]] = value
+
+    return matrix.tocsr()
